"""
Module with an implementation of ``ProbInput`` class.

The ``ProbInput`` class represents a probabilistic input model.
Each probabilistic input has a set of one-dimensional marginals each of which
is defined by an instance of the ``Marginal`` class.
"""

from __future__ import annotations

import numpy as np
import textwrap

from numpy.random._generator import Generator
from tabulate import tabulate
from typing import Any, List, Optional, Sequence

from .marginal import Marginal, FIELD_NAMES

__all__ = ["ProbInput"]


class ProbInput:
    """A class for multivariate input variables.

    Parameters
    ----------
    marginals : Union[List[Marginal], Tuple[Marginal, ...]]
        A list of one-dimensional marginals (univariate random variables).
    copulas : Any
        Copulas between univariate inputs that define dependence structure
        (currently not used).
    input_id: str, optional
        The ID of the probabilistic input. If not specified, the value is None.
    function_id: str, optional
        The ID of the function associated with the input. If not specified,
        the value is None.
    description: str, optional
        The short description regarding the input model.
    rng_seed : int, optional.
        The seed used to initialize the pseudo-random number generator.
        If not specified, the value is taken from the system entropy.
    """

    def __init__(
        self,
        marginals: Sequence[Marginal],
        copulas: Any = None,
        input_id: Optional[str] = None,
        function_id: Optional[str] = None,
        description: Optional[str] = None,
        rng_seed: Optional[int] = None,
    ):
        # Read-only properties
        self._marginals = marginals
        self._copulas = copulas
        # Attributes
        self.input_id = input_id
        self.function_id = function_id
        self.description = description
        # Other properties
        self._rng_seed = rng_seed
        self._rng: Optional[Generator] = None

    @property
    def input_dimension(self) -> int:
        """Return the number of constituents (random) input variables."""
        return len(self._marginals)

    @property
    def marginals(self) -> Sequence[Marginal]:
        """Return the sequence of Marginals that define the input variables."""
        return self._marginals

    @property
    def copulas(self) -> Any:
        """Return the underlying Copulas of the probabilistic input."""
        return self._copulas

    @property
    def rng_seed(self) -> Optional[int]:
        """Return the seed for RNG."""
        return self._rng_seed

    @rng_seed.setter
    def rng_seed(self, value: Optional[int]):
        """Set/reset the seed for RNG."""
        self.reset_rng(self._rng_seed)

    def transform_sample(self, xx: np.ndarray, other: ProbInput):
        """Transform a sample from the distribution to another."""
        # Make sure the dimensionality is consistent
        if self.input_dimension != other.input_dimension:
            raise ValueError(
                "The dimensionality of the two inputs are not consistent!"
            )

        xx_trans = xx.copy()
        if not self.copulas:
            # Independent inputs, transform marginal by marginal
            for idx_dim, (marginal_self, marginal_other) in enumerate(
                zip(self.marginals, other.marginals)
            ):
                xx_trans[:, idx_dim] = marginal_self.transform_sample(
                    xx[:, idx_dim], marginal_other
                )
        else:
            raise ValueError("Copulas are not currently supported!")

        return xx_trans

    def get_sample(self, sample_size: int = 1) -> np.ndarray:
        """Get a random sample from the distribution.

        Parameters
        ----------
        sample_size : int
            The number of sample points in the generated sample.

        Returns
        -------
        np.ndarray
            The generated sample in an :math:`N`-by-:math:`M` array
            where :math:`N` and :math:`M` are the sample size
            and the number of input dimensions, respectively.
        """
        if self._rng is None:  # pragma: no cover
            # Create a pseudo-random number generator (lazy evaluation)
            self._rng = np.random.default_rng(self.rng_seed)

        xx = self._rng.random((sample_size, self.input_dimension))
        if not self.copulas:
            # Transform the sample in [0, 1] to the domain of the distribution
            for idx_dim, marginal in enumerate(self.marginals):
                xx[:, idx_dim] = marginal.icdf(xx[:, idx_dim])
        else:
            raise ValueError("Copulas are not currently supported!")

        return xx

    def pdf(self, xx: np.ndarray) -> np.ndarray:
        """Get the PDF value of the distribution on a set of values.

        Parameters
        ----------
        xx : np.ndarray
            Sample values (realizations) from a distribution.

        Returns
        -------
        np.ndarray
            PDF values of the distribution on the sample values.
        """
        if not self.copulas:
            yy = np.empty(xx.shape)
            for i, marginal in enumerate(self.marginals):
                yy[:, i] = marginal.pdf(xx[:, i])
            # Use log-transform because of the smallness of numbers
            yy = np.exp(np.sum(np.log(yy), axis=1))
        else:
            raise ValueError("Copulas are not currently supported!")

        return yy

    def reset_rng(self, rng_seed: Optional[int]) -> None:
        """Reset the random number generator.

        Parameters
        ----------
        rng_seed : int, optional.
            The seed used to initialize the pseudo-random number generator.
            If not specified, the value is taken from the system entropy.
        """
        rng = np.random.default_rng(rng_seed)
        self._rng = rng
        self._rng_seed = rng_seed

    def __str__(self):
        """Return human-readable string representation of the instance."""
        if self.input_id is None or self.input_id == "":
            input_id = "-"
        else:
            input_id = self.input_id
        if self.function_id is None or self.function_id == "":
            function_id = "-"
        else:
            function_id = self.function_id
        table = f"Function ID     : {function_id}\n"
        table += f"Input ID        : {input_id}\n"
        table += f"Input Dimension : {self.input_dimension}\n"

        # Parse the description column
        if self.description is None or self.description == "":
            description = "-"
        else:
            desc = textwrap.wrap(self.description, width=57)
            # Pad new lines
            if len(desc) > 1:
                desc[1:] = ["                  " + line for line in desc[1:]]
            description = "\n".join(desc)
        table += f"Description     : {description}\n"
        table += "Marginals       :\n\n"

        # Get the header names
        header_names = [name.capitalize() for name in FIELD_NAMES]
        header_names.insert(0, "No.")

        # Get the values for each field as a list
        list_values = _get_values_as_list(self.marginals, FIELD_NAMES)

        table += tabulate(
            list_values,
            headers=header_names,
            stralign="center",
            disable_numparse=True,
        )

        # Temporary solution for independence copula
        copulas = "Independence" if self.copulas is None else self.copulas

        table += f"\n\nCopulas         : {copulas}"

        return table

    def __repr__(self):
        """Return the unambiguous string representation of the instance."""
        class_name = self.__class__.__name__
        # Get the value of the constructor arguments
        attrs = {
            "marginals": self.marginals,
            "copulas": self.copulas,
            "input_id": self.input_id,
            "function_id": self.function_id,
            "description": self.description,
            "rng_seed": self.rng_seed,
        }
        attrs_str = ", ".join(f"{k}={v!r}" for k, v in attrs.items())

        return f"{class_name}({attrs_str})"


def _get_values_as_list(
    univ_inputs: Sequence[Marginal],
    field_names: List[str],
) -> list:
    """Get the values from each field from a list of UnivariateInput

    Parameters
    ----------
    univ_inputs : Union[List[UnivariateInput], Tuple[UnivariateInput, ...]]
        A list or a tuple of UnivariateInput representing
        the marginal distribution.
    field_names : List[str]
        A list of field names from each UnivariateInput to access
        (and its value grabbed).

    Return
    ------
    list
        List of values.
    """
    list_values = []
    for i, marginal in enumerate(univ_inputs):
        values = [i + 1]
        for field_name in field_names:
            attr_value = getattr(marginal, field_name)
            if attr_value is None:
                attr_value = "-"
            values.append(attr_value)
        list_values.append(values)

    return list_values



"""
Utility module for the test_functions sub-package.
"""

import numpy as np


def deg2rad(xx: np.ndarray) -> np.ndarray:
    """Convert angles given in degree to radians.

    Parameters
    ----------
    xx : np.ndarray
        Angles in degree.

    Returns
    -------
    np.ndarray
        Angles in radians.
    """
    yy = np.pi / 180.0 * xx

    return yy


def lognorm2norm_std(lognormal_mean: float, lognormal_std: float) -> float:
    """Get the std of the underlying normal distribution of a lognormal dist.

    Parameters
    ----------
    lognormal_mean : float
        The mean of a lognormal distribution.
    lognormal_std : float
        The standard deviation of a lognormal distribution.

    Returns
    -------
    float
        The standard deviation of the underlying normal distribution.

    Notes
    -----
    - The logarithm of a lognormal distribution gives the underlying normal
      distribution.
    """
    normal_var = np.log(
        (lognormal_mean**2 + lognormal_std**2) / lognormal_mean**2
    )

    return np.sqrt(normal_var)


def lognorm2norm_mean(lognormal_mean: float, lognormal_std: float) -> float:
    """Get the mean of the underlying normal distribution of a lognormal dist.

    Parameters
    ----------
    lognormal_mean : float
        The mean of a lognormal distribution.
    lognormal_std : float
        The standard deviation of a lognormal distribution.

    Returns
    -------
    float
        The mean of the underlying normal distribution.

    Notes
    -----
    - The logarithm of a lognormal distribution gives the underlying normal
      distribution.
    """
    normal_std = lognorm2norm_std(lognormal_mean, lognormal_std)

    normal_mean = np.log(lognormal_mean) - normal_std**2 / 2.0

    return normal_mean


def gumbel_max_mu(mean: float, std: float) -> float:
    """Get the mu (location) parameter of a Gumbel (max.) distribution.

    Parameters
    ----------
    mean : float
        The mean of a Gumbel (max.) distribution.
    std : float
        The standard deviation of a Gumbel (max.) distribution.

    Returns
    -------
    float
        The mu (location) parameter of the Gumbel (max.) distribution.
    """
    beta = gumbel_max_beta(std)

    mu = mean - beta * np.euler_gamma

    return mu


def gumbel_max_beta(std: float) -> float:
    """Get the beta (scale) parameter of the Gumbel (max.) distribution.

    Parameters
    ----------
    std : float
        The standard deviation of a Gumbel (max.) distribution.

    Returns
    -------
    float
        The beta (scale) parameter of the Gumbel (max.) distribution.
    """
    beta = std * np.sqrt(6) / np.pi

    return beta



"""
Module with an implementation of the circular pipe crack reliability problem.

The two-dimensional reliability problem was introduced in [1] and used,
for instance, in [2].

The system under consideration is as a circular pipe with a circumferential
through-wall crack under a bending moment.
If the value of the performance function is negative,
then the system is in failed state.

References
----------
1. A. K. Verma, S. Ajit, and D. R. Karanki, “Structural Reliability,”
   in Reliability and Safety Engineering, in Springer Series
   in Reliability Engineering. London: Springer London, 2016, pp. 257–292.
   DOI: 10.1007/978-1-4471-6269-8_8
2. X. Li, C. Gong, L. Gu, W. Gao, Z. Jing, and H. Su, “A sequential surrogate
   method for reliability analysis based on radial basis function,”
   Structural Safety, vol. 73, pp. 42–53, 2018.
   DOI: 10.1016/j.strusafe.2018.02.005
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs, FunParamSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC


__all__ = ["CircularPipeCrack"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Verma2015": {
        "function_id": "CircularPipeCrack",
        "description": (
            "Input model for the circular pipe crack problem "
            "from Verma et al. (2015)"
        ),
        "marginals": [
            {
                "name": "sigma_f",
                "distribution": "normal",
                "parameters": [301.079, 14.78],
                "description": "flow stress [MNm]",
            },
            {
                "name": "theta",
                "distribution": "normal",
                "parameters": [0.503, 0.049],
                "description": "half crack angle [-]",
            },
        ],
        "copulas": None,
    },
}

AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Verman2016": {
        "function_id": "CircularPipeCrack",
        "description": (
            "Parameter set for the circular pipe crack reliability problem "
            "from Verma et al. (2016)"
        ),
        "declared_parameters": [
            {
                "keyword": "pipe_radius",
                "value": 3.377e-1,
                "type": float,
                "description": "Radius of the pipe [m]",
            },
            {
                "keyword": "pipe_thickness",
                "value": 3.377e-2,
                "type": float,
                "description": "Thickness of the pipe [m]",
            },
            {
                "keyword": "bending_moment",
                "value": 3.0,
                "type": float,
                "description": "Applied bending moment [Nm]",
            },
        ],
    },
}


def evaluate(
    xx: np.ndarray,
    pipe_radius: float,
    pipe_thickness: float,
    bending_moment: float,
) -> np.ndarray:
    """Evaluate the circular pipe crack reliability on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by N-by-2 arrays
        where N is the number of input values.
    pipe_radius : float
        The radius of the pipe in [m].
    pipe_thickness : float
        The thickness of the pipe in [m].
    bending_moment : float
        The applied bending moment in [Nm].

    Returns
    -------
    np.ndarray
        The performance evaluation of the circular pipe reliability.
        If negative, then the system is in failed state.
        The output is a one-dimensional array of length N.
    """
    # NOTE: Convert the flow stress from [MNm] to [Nm]
    yy = (
        4
        * pipe_thickness
        * xx[:, 0]
        * pipe_radius**2
        * (np.cos(xx[:, 1] / 2) - 0.5 * np.sin(xx[:, 1]))
        - bending_moment
    )

    return yy


class CircularPipeCrack(UQTestFunFixDimABC):
    """A concrete implementation of the circular pipe crack problem."""

    _tags = ["reliability"]
    _description = (
        "Circular pipe under bending moment from Verma et al. (2015)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Utility module for univariate distribution calculations.
"""

import numpy as np

from ....global_settings import ARRAY_FLOAT


def postprocess_icdf(
    xx: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Postprocess the computed ICDF values.

    The postprocessing ensures that the output is an array and always
    within the bounds of the distribution and

    Parameters
    ----------
    xx : ARRAY_FLOAT
        The raw ICDF values of a distribution computed either from a built-in
        function or a re-parameterization of the SciPy implementation.
    lower_bound : float
        The lower bound of the distribution. All values below this value will
        automatically be set to this value.
    upper_bound : float
        The upper bound of the distribution. All values above this value will
        automatically be set to this value.

    Return
    ------
    ARRAY_FLOAT
        The post-processed ICDF values.
    """
    # A scalar output
    if xx.ndim == 0:
        if xx < lower_bound:
            xx = np.asarray(lower_bound)
        if xx > upper_bound:
            xx = np.asarray(upper_bound)

        xx = np.asarray(xx)

    else:
        xx[xx < lower_bound] = lower_bound
        xx[xx > upper_bound] = upper_bound

    return xx


def verify_param_nums(given: int, expected: int, dist_name: str) -> None:
    """Verify the number of given parameters against the expected one.

    Parameters
    ----------
    given : int
        The given number of parameters of a distribution.
    expected : int
        The expected number of parameters of the distribution.
    dist_name : str
        The name of the distribution.
    Returns
    -------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If the given value is not the same as expected one.
    """
    if given != expected:
        raise ValueError(
            f"A {dist_name} distribution requires {expected} parameters! "
            f"Expected {expected}, got {given}."
        )



"""
Module with routines involving the uniform probability distribution.

The uniform distribution in UQTestFuns is parametrized by two parameters:
the lower and upper bounds.
"""

import numpy as np

from .utils import postprocess_icdf, verify_param_nums
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "uniform"

NUM_PARAMS = 2


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a uniform distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of the uniform distribution
        (i.e., lower and upper bounds).

    Returns
    ------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    if parameters[0] >= parameters[1]:
        raise ValueError(
            f"The lower bound {parameters[0]} "
            f"cannot be greater than the upper bound {parameters[1]}!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a uniform distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a uniform distribution.

    Returns
    -------
    float
        The lower bound of the uniform distribution.
    """
    lower_bound = float(parameters[0])

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a uniform distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a uniform distribution.

    Returns
    -------
    float
        The upper bound of the uniform distribution.
    """
    upper_bound = float(parameters[1])

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a uniform distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a uniform distribution.
    parameters : ARRAY_FLOAT
        Parameters of the uniform distribution.
    lower_bound: float
        Lower bound of the uniform distribution.
    upper_bound: float
        Upper bound of the distribution.

    Returns
    -------
    ARRAY_FLOAT
        The PDF values of the uniform distribution.

    Notes
    -----
    - The sample values ``xx`` themselves are not used in the computation of
      density value (it is, after all, a constant),
      but required nevertheless as the function is vectorized.
      Given a vector input, the function should return the PDF values of the
      same length as the input.
      Moreover, this signature must be consistent with the other distributions.
    - The values outside the bounds are set to 0.0.
    """
    yy = np.zeros(xx.shape)
    idx = np.logical_and(xx >= lower_bound, xx <= upper_bound)

    yy[idx] = 1 / (np.diff(parameters))

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of a uniform distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a uniform distribution.
    parameters : ARRAY_FLOAT
        Parameters of the uniform distribution.
    lower_bound : float
        Lower bound of the uniform distribution
    upper_bound : float
        Upper bound of the uniform distribution.

    Returns
    -------
    ARRAY_FLOAT
        The CDF values of the uniform distribution.

    Notes
    -----
    - The CDF values for sample values below the lower bound are set to 0.0,
      and for sample values above the upper bound are set to 1.0.
    """
    yy = np.empty(xx.shape)
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_rest] = (xx[idx_rest] - parameters[0]) / np.diff(parameters)

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a uniform distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) in the [0, 1] domain.
    parameters : ARRAY_FLOAT
        Parameters of a uniform distribution.
    lower_bound : float
        Lower bound of the uniform distribution.
    upper_bound : float
        Upper bound of the uniform distribution.
        This parameter is not used but must appear for interface consistency.

    Returns
    -------
    np.ndarray
        Transformed values in the domain of the uniform distribution.
    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN.
    """
    xx[xx < 0.0] = np.nan
    xx[xx > 1.0] = np.nan

    # Compute the ICDF
    yy = lower_bound + np.diff(parameters) * xx

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Module with an implementation of the hyper-sphere bound reliability problem.

The reliability problem is a two-dimensional function as described in
Li et al. (2018).

References
----------

1. X. Li, C. Gong, L. Gu, W. Gao, Z. Jing, and H. Su, “A sequential surrogate
   method for reliability analysis based on radial basis function,”
   Structural Safety, vol. 73, pp. 42–53, 2018.
   DOI: 10.1016/j.strusafe.2018.02.005.
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["HyperSphere"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Li2018": {
        "function_id": "Li2018",
        "description": (
            "Input model for the hyper-sphere reliability problem "
            "from Li et al. (2018)"
        ),
        "marginals": [
            {
                "name": "X1",
                "distribution": "normal",
                "parameters": [0.5, 0.2],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "normal",
                "parameters": [0.5, 0.2],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the hyper-sphere performance function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        1-Dimensional input values given by an N-by-1 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = 1 - xx[:, 0] ** 3 - xx[:, 1] ** 3

    return yy


class HyperSphere(UQTestFunFixDimABC):
    """A concrete implementation of the hyper-sphere reliability problem."""

    _tags = ["reliability"]
    _description = (
        "Hyper-sphere bound reliability problem from Li et al. (2018)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_dimension = 2

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module providing high-level helper functions.

Notes
-----
- High-level functions typically contain many guardrails against invalidity
  of the input arguments.
"""

from tabulate import tabulate as tbl  # 'tabulate' is used as a parameter name
from .utils import get_available_classes, SUPPORTED_TAGS
from . import test_functions
from typing import List, Optional, Union

from .core import UQTestFunABC

__all__ = ["list_functions"]


HEADER: dict = {
    "input_dim": {
        "header_name": "# Input",
        "colalign": "center",
        "maxcolwidth": None,
    },
    "output_dim": {
        "header_name": "# Output",
        "colalign": "center",
        "maxcolwidth": None,
    },
    "parameterized": {
        "header_name": "Param.",
        "colalign": "center",
        "maxcolwidth": None,
    },
    "tags": {
        "header_name": "Application",
        "colalign": "center",
        "maxcolwidth": 20,
    },
    "description": {
        "header_name": "Description",
        "colalign": "left",
        "maxcolwidth": 30,
    },
}


def list_functions(
    input_dimension: Optional[Union[str, int]] = None,
    tag: Optional[str] = None,
    output_dimension: Optional[int] = None,
    parameterized: Optional[bool] = None,
    tabulate: bool = True,
    tablefmt: str = "grid",
) -> Optional[Union[List[UQTestFunABC], str]]:
    """List of all the available functions.

    Parameters
    ----------
    input_dimension : Optional[Union[str, int]]
        Filter based on the number of input dimension.
        For variable dimension (i.e., M-dimensional test functions),
        use the string "M".
    tag : Optional[str]
        Filter based on the application tag.
        Supported tags: "metamodeling", "sensitivity", "optimization",
        "reliability".
    output_dimension : Optional[Union[str, int]]
        Filter based on the number of output dimension.
    parameterized : bool, optional
        Filter based on whether the test function is parameterized.
    tabulate : bool, optional
        The flag whether to print a table on the console or a list
        of the available functions (each in fully-qualified class name).
    tablefmt : str, optional
        Format of the table output; use "html" to return a table in HTML
        format nicely rendered in Jupyter notebook.

    Returns
    -------
    Optional[List[UQTestFunABC]]
        Either a tabulated view of the list of available functions,
        or a list of fully-qualified class name (each is callable).
        The function may return None if after filtering there is no entry.

    Notes
    -----
    - When both ``input_dimension`` and ``tag`` are specified the results are
      intersected. Entries that satisfy both are returned.
    """

    # --- Parse input arguments
    _verify_input_args(
        input_dimension, tag, output_dimension, parameterized, tabulate
    )

    # --- Parse the module-level data
    data = _parse_modules_data(test_functions)

    # --- Filter based on the input dimension
    data = _filter_on_input_dim(data, input_dimension)

    # --- Filter based on the output dimension
    data = _filter_on_output_dim(data, output_dimension)

    # --- Filter based on parameterization
    data = _filter_on_parameterized(data, parameterized)

    # --- Filter based on the tags
    data = _filter_on_tag(data, tag)

    # --- When asked, immediately return all the fully-qualified class name
    if not tabulate:
        constructors = [
            data[k]["full_path"] for k in sorted(list(data.keys()))
        ]
        return constructors

    # --- Get the arguments for tabulate
    print_attribs, header_names, colalign, maxcolwidth = _get_table_formatting(
        input_dimension,
        tag,
        output_dimension,
        parameterized,
    )

    values = _create_list_values(data, print_attribs)

    if len(values) == 0:
        return None

    if tablefmt == "html":
        colalign[-1] = "center"
        table = tbl(
            values,
            headers=header_names,
            tablefmt=tablefmt,
            colalign=colalign,
            maxcolwidths=maxcolwidth,
        )
        return table
    else:
        table = tbl(
            values,
            headers=header_names,
            tablefmt=tablefmt,
            colalign=colalign,
            maxcolwidths=maxcolwidth,
        )
        print(table)

    return None


def _verify_input_args(
    input_dimension: Optional[Union[str, int]] = None,
    tag: Optional[str] = None,
    output_dimension: Optional[int] = None,
    parameterized: Optional[bool] = None,
    tabulate: bool = True,
) -> None:
    """Verify the input arguments.

    Parameters
    ----------
    input_dimension : Optional[Union[str, int]]
        The number of input dimension to filter the list of test functions.
        For variable dimension (i.e., M-dimensional test functions),
        use the string "M".
    tag : Optional[str]
        The application tag to filter the list of test functions.
        Supported tags: "metamodeling", "sensitivity", "optimization",
        "reliability".
    output_dimension : int, optional
        The number of output dimension to filter the list of test functions.
    parameterized : bool, optional
        The flag based on whether the test function is parameterized to filter
        the list of test functions.
    tabulate : bool, optional
        The flag whether to print a table on the console or a list
        of the available functions (each in fully-qualified class name).

    Raises
    ------
    ValueError
        If ``input_dimension`` is not a positive integer or the string "M".
        If ``tag`` is not one of the supported tags.
        If ``output_dimension`` is not a positive integer.
    TypeError
        If ``input_dimension`` is not either an integer, string, or NoneType.
        If ``tag`` is not a string.
        If ``output_dimension`` is not an integer, string, or NoneType.
        If ``parameterized`` is not a bool.
        If ``tabulate`` is not a bool.
    """
    # --- Parse 'input_dimension'
    if not isinstance(input_dimension, (int, str, type(None))):
        raise TypeError(
            f"Invalid type for input dimension! "
            f"Expected either an integer or a string. "
            f"Got instead {type(input_dimension)}."
        )
    if input_dimension is not None and isinstance(input_dimension, str):
        if input_dimension.lower() != "m":
            raise ValueError(
                f"Invalid value ({input_dimension}) for input dimension! "
                f"Either a positive integer or 'M' to indicate "
                f"a variable-dimension test function."
            )
    if input_dimension is not None and isinstance(input_dimension, int):
        if input_dimension <= 0:
            raise ValueError(
                f"Invalid value ({input_dimension}) for input dimension! "
                f"Either a positive integer or 'M' to indicate "
                f"a variable-dimension test function."
            )

    # --- Parse 'tag'
    if not isinstance(tag, (str, type(None))):
        raise TypeError(f"Tag argument must be of str type! Got {type(tag)}.")
    if tag is not None and tag not in SUPPORTED_TAGS:
        raise ValueError(
            f"Tag {tag!r} is not supported. Use one of {SUPPORTED_TAGS}!"
        )

    # --- Parse 'input_dimension'
    if not isinstance(output_dimension, (int, type(None))):
        raise TypeError(
            f"Invalid type for output dimension! "
            f"Expected either an integer or a string. "
            f"Got instead {type(output_dimension)}."
        )
    if output_dimension is not None:
        if output_dimension <= 0:
            raise ValueError(
                f"Invalid value ({output_dimension}) for output dimension! "
                f"Must be a positive integer."
            )

    # --- Parse 'parameterized'
    if not isinstance(parameterized, (bool, type(None))):
        raise TypeError(
            f"'parameterized' argument must be of bool type! "
            f"Got {type(parameterized)}."
        )

    # --- Parse 'tabulate'
    if not isinstance(tabulate, (bool, type(None))):
        raise TypeError(
            f"'tabulate' argument must be of bool type! Got {type(tabulate)}."
        )


def _filter_on_input_dim(data, input_dimension):
    """Filter the dictionary of test functions data based on the input dim."""
    if input_dimension is not None:
        if isinstance(input_dimension, str):
            input_dimension = input_dimension.upper()
        data = {
            k: v for k, v in data.items() if v["input_dim"] == input_dimension
        }

    return data


def _filter_on_output_dim(data, output_dimension):
    """Filter the dictionary of test functions data based on the output dim."""
    if output_dimension is not None:
        data = {
            k: v
            for k, v in data.items()
            if v["output_dim"] == output_dimension
        }

    return data


def _filter_on_parameterized(data, parameterized):
    """Filter the dictionary of test functions data based on the param."""
    if parameterized is not None:
        data = {
            k: v
            for k, v in data.items()
            if v["parameterized"] is parameterized
        }

    return data


def _filter_on_tag(data, tag):
    """Filter the dictionary of test functions data based on the tag."""
    if tag is not None:
        data = {k: v for k, v in data.items() if tag in v["tags"]}

    return data


def _create_list_values(data, print_attribs):
    """Get the selected values from dictionary test functions data."""
    values = []
    for i, function_name in enumerate(sorted(list(data.keys()))):
        values_tmp = [i + 1, data[function_name]["constructor"]]
        for print_attrib in print_attribs:
            values_tmp.append(data[function_name][print_attrib])
        values.append(values_tmp)

    return values


def _get_table_formatting(
    input_dimension,
    tag,
    output_dimension,
    parameterized,
):
    """Get the table formatting as arguments to 'tabulate()'."""
    header_names: List[str] = ["No.", "Constructor"]
    colalign: List[str] = ["center", "center"]
    maxcolwidth: List[Optional[int]] = [None, None]
    print_attribs: List[str] = []

    # The printed attribute should appear in this order
    if input_dimension is None:
        print_attribs.append("input_dim")

    if output_dimension is None:
        print_attribs.append("output_dim")

    if parameterized is None:
        print_attribs.append("parameterized")

    if tag is None:
        print_attribs.append("tags")

    # --- Always get description
    print_attribs.append("description")

    # --- Get the arguments for tabulate
    for print_attrib in print_attribs:
        header_names.append(HEADER[print_attrib]["header_name"])
        colalign.append(HEADER[print_attrib]["colalign"])
        maxcolwidth.append(HEADER[print_attrib]["maxcolwidth"])

    return print_attribs, header_names, colalign, maxcolwidth


def _parse_modules_data(package):
    available_classes = get_available_classes(package)

    data = {}

    for available_class, class_path in available_classes:

        # Create an instance of test function to parse its properties
        instance: UQTestFunABC = class_path()

        # Get the dimension
        if instance.variable_dimension:
            input_dimension = "M"
        else:
            input_dimension = str(instance.input_dimension)

        data[available_class] = {
            "constructor": available_class + "()",
            "input_dim": input_dimension,
            "output_dim": instance.output_dimension,
            "parameterized": True if instance.parameters else False,
            "tags": ", ".join(instance.tags),
            "description": instance.description,
            "full_path": class_path,
        }

    return data



"""
Module with an implementation of the ``Marginal`` class.

The ``Marginal`` class represents a one-dimensional marginal random variable
(i.e., univariate random variable).
Each random variable has a (parametric) probability distribution.
"""

from __future__ import annotations

import numpy as np

from numpy.random._generator import Generator
from numpy.typing import ArrayLike
from dataclasses import dataclass, field
from typing import Optional, Union

from .utils import (
    verify_distribution,
    get_distribution_bounds,
    verify_parameters,
    get_pdf_values,
    get_cdf_values,
    get_icdf_values,
)
from ...global_settings import ARRAY_FLOAT

__all__ = ["Marginal"]

# Ordered field names for printing purpose
FIELD_NAMES = ["name", "distribution", "parameters", "description"]


@dataclass(frozen=True)
class Marginal:
    """A class for one-dimensional marginal random variables.

    Parameters
    ----------
    distribution : str
        The type of the probability distribution.
    parameters : ArrayLike
        The parameters of the chosen probability distribution
    name : str, optional
        The name of the random variable
    description : str, optional
        The short text description of the random variable
    rng_seed : int, optional.
        The seed used to initialize the pseudo-random number generator.
        If not specified, the value is taken from the system entropy.

    Attributes
    ----------
    lower : float
        The lower bound of the distribution
    upper : float
        The upper bound of the distribution
    _rng : Generator
        The default pseudo-random number generator of NumPy.
        The generator is only created if or when needed (e.g., generating
        a random sample from the distribution).
    """

    distribution: str
    parameters: ArrayLike
    name: Optional[str] = None
    description: Optional[str] = None
    lower: float = field(init=False, repr=False)
    upper: float = field(init=False, repr=False)
    rng_seed: Optional[int] = field(default=None, repr=False)
    _rng: Optional[Generator] = field(init=False, default=None, repr=False)

    def __post_init__(self) -> None:
        # Because frozen=True, post init must access self via setattr
        # Make sure the distribution is lower-case
        object.__setattr__(self, "distribution", self.distribution.lower())

        # Convert parameters to a numpy array
        object.__setattr__(self, "parameters", np.array(self.parameters))

        # Verify the selected univariate distribution type
        verify_distribution(self.distribution)

        # Verify the value of the parameters
        verify_parameters(self.distribution, self.parameters)

        # Get and set the lower and upper bounds
        _lower, _upper = get_distribution_bounds(
            self.distribution, self.parameters
        )
        object.__setattr__(self, "lower", _lower)
        object.__setattr__(self, "upper", _upper)

    def transform_sample(
        self,
        xx: Union[float, np.ndarray],
        other: Marginal,
    ) -> np.ndarray:
        """Transform a sample from a given distribution to another."""
        if not isinstance(other, Marginal):
            raise TypeError("Other instance must be of Marginal type!")

        xx_trans = self.cdf(xx)

        return get_icdf_values(
            xx_trans,
            other.distribution,
            other.parameters,
            other.lower,
            other.upper,
        )

    def get_sample(self, sample_size: int = 1) -> np.ndarray:
        """Get a random sample from the distribution."""
        if self._rng is None:  # pragma: no cover
            # Create a pseudo-random number generator (lazy evaluation)
            rng = np.random.default_rng(self.rng_seed)
            object.__setattr__(self, "_rng", rng)

        xx = self._rng.random(sample_size)  # type: ignore

        return get_icdf_values(
            xx, self.distribution, self.parameters, self.lower, self.upper
        )

    def reset_rng(self, rng_seed: Optional[int]) -> None:
        """Reset the random number generator.

        Parameters
        ----------
        rng_seed : int, optional.
            The seed used to initialize the pseudo-random number generator.
            If not specified, the value is taken from the system entropy.
        """
        rng = np.random.default_rng(rng_seed)
        object.__setattr__(self, "_rng", rng)
        object.__setattr__(self, "rng_seed", rng_seed)

    def pdf(self, xx: Union[float, np.ndarray]) -> ARRAY_FLOAT:
        """Compute the PDF of the distribution on a set of values."""
        # TODO: check if you put a scalar inside
        # Convert input to an np.array
        xx = np.asarray(xx)

        return get_pdf_values(
            xx, self.distribution, self.parameters, self.lower, self.upper
        )

    def cdf(self, xx: Union[float, np.ndarray]) -> ARRAY_FLOAT:
        """Compute the CDF of the distribution on a set of values.

        The function transforms the sample values in the domain
        of the distribution to the [0, 1] domain.
        """
        # TODO: check if you put a scalar inside
        # Convert input to an np.array
        xx = np.asarray(xx)

        return get_cdf_values(
            xx, self.distribution, self.parameters, self.lower, self.upper
        )

    def icdf(self, xx: Union[float, np.ndarray]) -> ARRAY_FLOAT:
        """Compute the inverse CDF of the distribution on a set of values.

        The function transforms values in the [0,1] domain to the domain
        of the distribution.
        """
        # TODO: verify that the input is in [0, 1]
        xx = np.asarray(xx)

        return get_icdf_values(
            xx, self.distribution, self.parameters, self.lower, self.upper
        )



"""
Module with an implementation of the 2D function from Webster et al. (1996).

The two-dimensional function is polynomial with random input variables.
It was introduced in [1] and used to illustrate the construction of polynomial
chaos expansion metamodel.

References
----------
1. M. Webster, M. A. Tatang, and G. J. McRae, “Application of the probabilistic
   collocation method for an uncertainty analysis of a simple ocean model,”
   Massachusetts Institute of Technology, Cambridge, MA,
   Joint Program Report Series 4, 1996.
   [Online]. Available: http://globalchange.mit.edu/publication/15670
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Webster2D"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Webster1996": {
        "function_id": "Webster2D",
        "description": (
            "Input specification for the 2D function "
            "from Webster et al. (1996)"
        ),
        "marginals": [
            {
                "name": "A",
                "distribution": "uniform",
                "parameters": [1.0, 10.0],
                "description": None,
            },
            {
                "name": "B",
                "distribution": "normal",
                "parameters": [2.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray):
    """Evaluate the 2D Webster function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the 2D Webster function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = xx[:, 0] ** 2 + xx[:, 1] ** 3

    return yy


class Webster2D(UQTestFunFixDimABC):
    """A concrete implementation of the function from Webster et al. (1996)."""

    _tags = ["metamodeling"]
    _description = "2D polynomial function from Webster et al. (1996)."
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
The core subpackage of uqtestfuns.
"""

from .parameters import FunParams
from .prob_input.marginal import Marginal
from .prob_input.probabilistic_input import ProbInput
from .uqtestfun_abc import (
    UQTestFunBareABC,
    UQTestFunABC,
    UQTestFunFixDimABC,
    UQTestFunVarDimABC,
)
from .uqtestfun import UQTestFun

__all__ = [
    "Marginal",
    "ProbInput",
    "FunParams",
    "UQTestFunBareABC",
    "UQTestFunABC",
    "UQTestFunFixDimABC",
    "UQTestFunVarDimABC",
    "UQTestFun",
]



"""
This module provides abstract base classes for defining a test function class.
"""

import abc
from abc import ABC

import numpy as np

from copy import deepcopy
from inspect import signature
from typing import Callable, cast, List, Optional, Type

from .prob_input.marginal import Marginal
from .prob_input.probabilistic_input import ProbInput
from .parameters import FunParams
from .custom_typing import (
    ProbInputSpecs,
    ProbInputArgs,
    FunParamSpecs,
    FunParamsArgs,
)
from .utils import create_canonical_uniform_input

__all__ = [
    "UQTestFunBareABC",
    "UQTestFunABC",
    "UQTestFunFixDimABC",
    "UQTestFunVarDimABC",
]

CLASS_HIDDEN_ATTRIBUTES = [
    "_tags",
    "_description",
    "_available_inputs",
    "_available_parameters",
]

DEFAULT_DIMENSION = 2


class classproperty(property):  # type: ignore
    """Decorator w/ descriptor to get and set class-level attributes."""

    def __get__(self, owner_self, owner_cls):  # type: ignore
        return self.fget(owner_cls)  # type: ignore

    def __set__(self, owner_self, owner_cls):  # pragma: no cover
        raise AttributeError("can't set attribute")


class UQTestFunBareABC(abc.ABC):
    """An abstract class for a bare UQ test functions.

    Parameters
    ----------
    prob_input : ProbInput
        The probabilistic input model of the UQ test function.
    parameters : Any, optional
        A set of parameters. By default, it is None.
    function_id : str, optional
        The ID of the UQ test function. By default, it is None.

    Notes
    -----
    - A bare UQ test function only includes the evaluation function,
      probabilistic input model, parameters, and a (optional) ID.
    """

    def __init__(
        self,
        prob_input: ProbInput,
        parameters: FunParams,
        function_id: Optional[str] = None,
    ):
        self.prob_input = prob_input
        self._parameters = parameters
        self._function_id = function_id
        self._output_dimension: Optional[int] = None

    @property
    def prob_input(self) -> ProbInput:
        """The probabilistic input model of the UQ test function."""
        return self._prob_input

    @prob_input.setter
    def prob_input(self, value: ProbInput):
        """The setter for probabilistic input model of the UQ test function."""
        if isinstance(value, ProbInput):
            self._prob_input = value
        else:
            raise TypeError(
                f"Probabilistic input model must be of "
                f"a 'ProbInput' type! Got instead {type(value)}."
            )

    @property
    def parameters(self) -> FunParams:
        """The parameters of the UQ test function."""
        return self._parameters

    @parameters.setter
    def parameters(self, value: FunParams):
        """The setter for the parameters of the test function."""
        if not isinstance(value, FunParams):
            raise TypeError(
                f"Expected a 'FunParams' type! Got instead {type(value)}"
            )
        self._parameters = value

    @property
    def function_id(self) -> Optional[str]:
        """The ID of the UQ test function."""
        return self._function_id

    @property
    def input_dimension(self) -> int:
        """The input dimension of the UQ test function."""
        return self.prob_input.input_dimension

    @property
    def output_dimension(self) -> int:
        if self._output_dimension is None:
            xx = self.prob_input.get_sample()
            yy = self(xx)
            if yy.ndim == 1:
                self._output_dimension = 1
            elif yy.ndim == 2:
                self._output_dimension = yy.shape[1]
            else:
                self._output_dimension = yy.shape[1:]
            return self._output_dimension

        return self._output_dimension

    def transform_sample(
        self,
        xx: np.ndarray,
        min_value=-1.0,
        max_value=1.0,
    ) -> np.ndarray:
        """Transform sample values from a unif. domain to the func. domain

        Parameters
        ----------
        xx : np.ndarray
            Sampled input values (realizations) in a uniform bounded domain.
            By default, the uniform domain is [-1.0, 1.0].
        min_value : float, optional
            Minimum value of the uniform domain. Default value is -1.0.
        max_value : float, optional
            Maximum value of the uniform domain. Default value is 1.0.

        Returns
        -------
        np.ndarray
            Transformed sampled values from the specified uniform domain to
            the domain of the function as defined the `input` property.
        """

        # Verify the uniform bounds
        assert min_value < max_value, (
            f"min. value ({min_value}) must be "
            f"smaller than max. value ({max_value})!"
        )
        # Verify the sampled input
        _verify_sample_shape(xx, self.input_dimension)
        _verify_sample_domain(xx, min_value=min_value, max_value=max_value)

        # Create an input in the canonical uniform domain
        uniform_input = create_canonical_uniform_input(
            self.input_dimension, min_value, max_value
        )

        # Transform the sampled value to the function domain
        xx_trans = uniform_input.transform_sample(xx, other=self.prob_input)

        return xx_trans

    def __str__(self):
        out = (
            f"Function ID      : {self.function_id}\n"
            f"Input Dimension  : {self.input_dimension}\n"
            f"Output Dimension : {self.output_dimension}\n"
            f"Parameterized    : {bool(self.parameters)}"
        )

        return out

    def __call__(self, xx):
        """Evaluation of the test function by calling the instance."""
        # Verify the shape of the input
        _verify_sample_shape(xx, self.input_dimension)

        # Verify the domain of the input
        for dim_idx in range(self.input_dimension):
            lb = self.prob_input.marginals[dim_idx].lower
            ub = self.prob_input.marginals[dim_idx].upper
            _verify_sample_domain(xx[:, dim_idx], min_value=lb, max_value=ub)

        return self._eval(xx)

    @staticmethod
    @abc.abstractmethod
    def evaluate(xx: np.ndarray, **kwargs) -> np.ndarray:
        """Abstract method for the implementation of the UQ test function."""
        pass

    def _eval(self, xx) -> np.ndarray:
        """Actual computation is delegated to evaluate()."""
        return self.__class__.evaluate(xx, **self.parameters.as_dict())


class UQTestFunABC(UQTestFunBareABC, ABC):
    """An abstract class for (published) UQ test functions.

    Parameters
    ----------
    input_dimension : int, optional
        The input dimension of the UQ test function.
        This is used only when the function supports variable dimension;
        otherwise, if specified, an exception is raised.
        In the case of functions with variable dimension, the default dimension
        is set to 2.
        This is a keyword only parameter.
    input_id : str, optional
        The selection of probabilistic input model; this is used when there are
        multiple input specifications in the literature.
        This is a keyword only parameter.
    parameters_id : str, optional
        The selection of parameters set; this is used when there are multiple
        sets of parameters available in the literature.
        This is a keyword only parameter.
    function_id : str, optional
        The ID of the UQ test function.
        If not given, `None` is used as the function ID.
        This is a keyword only parameter.

    Notes
    -----
    - A published UQ test function includes a couple of additional metadata,
      namely tags and description.

    Raises
    ------
    KeyError
        If selection is not in the list of available inputs and parameters.
    TypeError
        If input dimension is specified for a UQ test function with
        a fixed dimension.
    """

    _default_input_id: Optional[str] = None
    _default_parameters_id: Optional[str] = None

    @classproperty
    def tags(cls) -> List[str]:
        """Tags to classify different UQ test functions."""
        return cls._tags  # type: ignore

    @classproperty
    def available_inputs(cls) -> ProbInputSpecs:
        """All the keys to the available probabilistic input specifications."""
        return cls._available_inputs  # type: ignore

    @classproperty
    def default_input_id(cls) -> Optional[str]:
        """The key to the default probabilistic input specification."""
        return cls._default_input_id  # type: ignore

    @classproperty
    def available_parameters(cls) -> Optional[FunParamSpecs]:
        """All the keys to the available set of parameter values."""
        return cls._available_parameters  # type: ignore

    @classproperty
    def default_parameters_id(cls) -> Optional[str]:
        """The key to the default set of parameters."""
        return cls._default_parameters_id  # type: ignore

    @classproperty
    def description(cls) -> Optional[str]:
        """Short description of the UQ test function."""
        return cls._description  # type: ignore

    @property
    @abc.abstractmethod
    def variable_dimension(self):
        pass

    def __str__(self):
        if self.variable_dimension:
            input_dimension = f"{self.input_dimension} (variable)"
        else:
            input_dimension = f"{self.input_dimension} (fixed)"
        tags = ", ".join(self.tags)
        out = (
            f"Function ID      : {self.function_id}\n"
            f"Input Dimension  : {input_dimension}\n"
            f"Output Dimension : {self.output_dimension}\n"
            f"Parameterized    : {bool(self.parameters)}\n"
            f"Description      : {self.description}\n"
            f"Applications     : {tags}"
        )

        return out

    def _verify_input_id(self, input_id: Optional[str]) -> str:
        """Verify the 'input_id' argument.

        Parameters
        ----------
        input_id : Optional[str]
            The ID of the probabilistic input specification.

        Returns
        -------
        str
            The verified ID of the probabilistic input specification.
        """
        # --- Verify the input
        if input_id is None:
            input_id = self.default_input_id
            input_id = cast(str, input_id)

        available_inputs = self.available_inputs
        if input_id not in available_inputs:
            raise KeyError(
                f"Input ID {input_id} is not in the available "
                f"specifications {list(available_inputs.keys())}"
            )

        return input_id

    def _verify_parameters_id(self, parameters_id: Optional[str]) -> str:
        """Verify the 'parameters_id' argument.

        Parameters
        ----------
        parameters_id : Optional[str]
            The ID of the function parameters specification.

        Returns
        -------
        str
            The verified ID of the function parameters specification.
        """
        available_parameters = self.available_parameters

        if self.available_parameters is None and parameters_id is not None:
            raise ValueError("No parameters are available")

        if self.available_parameters is None:
            return ""

        if parameters_id is None:
            parameters_id = self.default_parameters_id
            parameters_id = cast(str, parameters_id)

        if parameters_id not in available_parameters:
            raise KeyError(
                f"Input ID {parameters_id} is not in the available "
                f"specifications {list(available_parameters.keys())}"
            )

        return parameters_id


class UQTestFunFixDimABC(UQTestFunABC, ABC):
    """An abstract class for (published) fixed-dimension UQ test functions.

    Parameters
    ----------
    input_id : str, optional
        The selection of probabilistic input model; this is used when there are
        multiple input specifications in the literature.
        This is a keyword only parameter.
    parameters_id : str, optional
        The selection of parameters set; this is used when there are multiple
        sets of parameters available in the literature.
        This is a keyword only parameter.
    function_id : str, optional
        The ID of the UQ test function.
        If not given, `None` is used as the function ID.
        This is a keyword only parameter.

    Notes
    -----
    - A published UQ test function includes a couple of additional metadata,
      namely tags and description.

    Raises
    ------
    KeyError
        If selection is not in the list of available inputs and parameters.
    TypeError
        If input dimension is specified for a UQ test function with
        a fixed dimension.
    """

    def __init__(
        self,
        *,
        input_id: Optional[str] = None,
        parameters_id: Optional[str] = None,
        function_id: Optional[str] = None,
    ):
        # --- Create a probabilistic input model
        input_id = self._verify_input_id(input_id)
        prob_input = self._create_prob_input(input_id)

        # --- Create a set of function parameters
        parameters_id = self._verify_parameters_id(parameters_id)
        fun_params = self._create_fun_params(parameters_id)

        # --- Process the default ID
        if function_id is None:
            function_id = self.__class__.__name__

        # --- Initialize the parent class
        super().__init__(
            prob_input=prob_input,
            parameters=fun_params,
            function_id=function_id,
        )

    def __init_subclass__(cls, **kwargs):
        """Verify if concrete class has all the required hidden attributes."""
        _init_subclass(cls)

    @property
    def variable_dimension(self) -> bool:
        """Return ``False`` due to fixed dimension function."""
        return False

    def _create_prob_input(self, input_id: str) -> ProbInput:
        """Create an instance of probabilistic input.

        Parameters
        ----------
        input_id : str
            The ID of the available probabilistic input models.

        Returns
        -------
        ProbInput
            The probabilistic input model based on the selected ID.
        """
        # Get the input (copy to avoid mutation)
        raw_data = deepcopy(self.available_inputs[input_id])

        # Process the marginals
        marginals = []
        for marginal in raw_data["marginals"]:
            marginals.append(Marginal(**marginal))

        # Recast the type to satisfy type checker
        input_data = cast(ProbInputArgs, raw_data)
        input_data["input_id"] = input_id
        input_data["marginals"] = marginals

        return ProbInput(**input_data)

    def _create_fun_params(self, parameters_id: Optional[str]) -> FunParams:
        """Create an instance of function parameters.

        Parameters
        ----------
        parameters_id : str
            The ID of the available function parameters.

        Returns
        -------
        FunParams
            The set of function parameters based on the selected ID.
        """
        if self.available_parameters is None:
            return FunParams()

        # Must be deep-copied due to modification
        param_data = deepcopy(self.available_parameters[parameters_id])

        # Prepare the dictionary as input
        param_data = cast(FunParamsArgs, param_data)  # for type checker
        param_data["parameter_id"] = parameters_id

        return FunParams(**param_data)


class UQTestFunVarDimABC(UQTestFunABC, ABC):
    """An abstract class for (published) variable-dimension UQ test functions.

    Parameters
    ----------
    input_dimension : int, optional
        The input dimension of the UQ test function.
        This is used only when the function supports variable dimension;
        otherwise, if specified, an exception is raised.
        In the case of functions with variable dimension, the default dimension
        is set to 2.
    input_id : str, optional
        The selection of probabilistic input model; this is used when there are
        multiple input specifications in the literature.
        This is a keyword only parameter.
    parameters_id : str, optional
        The selection of parameters set; this is used when there are multiple
        sets of parameters available in the literature.
        This is a keyword only parameter.
    function_id : str, optional
        The ID of the UQ test function.
        If not given, `None` is used as the function ID.
        This is a keyword only parameter.

    Notes
    -----
    - A published UQ test function includes a couple of additional metadata,
      namely tags and description.

    Raises
    ------
    KeyError
        If selection is not in the list of available inputs and parameters.
    TypeError
        If input dimension is specified for a UQ test function with
        a fixed dimension.
    """

    def __init__(
        self,
        input_dimension: int = DEFAULT_DIMENSION,
        *,
        input_id: Optional[str] = None,
        parameters_id: Optional[str] = None,
        function_id: Optional[str] = None,
    ):
        # --- Create a probabilistic input model
        input_id = self._verify_input_id(input_id)
        prob_input = self._create_prob_input(input_id, input_dimension)

        # --- Select the parameters set, when applicable
        parameters_id = self._verify_parameters_id(parameters_id)
        fun_params = self._create_fun_params(parameters_id, input_dimension)

        # --- Process the default ID
        if function_id is None:
            function_id = self.__class__.__name__

        # --- Initialize the parent class
        super().__init__(
            prob_input=prob_input,
            parameters=fun_params,
            function_id=function_id,
        )

    def __init_subclass__(cls, **kwargs):
        """Verify if concrete class has all the required hidden attributes."""
        _init_subclass(cls)

    @property
    def variable_dimension(self) -> bool:
        return True

    def _create_prob_input(self, input_id: str, input_dim: int) -> ProbInput:
        """Create an instance of probabilistic input model.

        Parameters
        ----------
        input_id : str
            The ID of the available probabilistic input models.
        input_dim : int
            The input dimension of the UQ test function.

        Returns
        -------
        ProbInput
            The probabilistic input model based on the selected ID.
        """

        # Get the input
        raw_data = deepcopy(self.available_inputs[input_id])

        # Process the marginals
        marginal = raw_data["marginals"]
        if callable(marginal):
            # Marginal specification is given as a function
            marginals = marginal(input_dim)
            marginals = [Marginal(**marginal) for marginal in marginals]
        else:
            # Marginal specification will be spawned up to the # of dimension
            marginals = []
            for i in range(input_dim):
                marginal = raw_data["marginals"][0].copy()
                marginal["name"] = f"{marginal['name']}{i}"
                marginals.append(Marginal(**marginal))

        # Recast the type to satisfy type checker
        input_data = cast(ProbInputArgs, raw_data)
        input_data["input_id"] = input_id
        input_data["marginals"] = marginals

        return ProbInput(**input_data)

    def _create_fun_params(
        self,
        parameters_id: Optional[str],
        input_dim: int,
    ) -> FunParams:
        """Create an instance of function parameters.

        Parameters
        ----------
        parameters_id : str
            The ID of the available function parameters.
        input_dim : int
            The input dimension of the UQ test function.

        Returns
        -------
        FunParams
            The set of function parameters based on the selected ID.
        """

        if self.available_parameters is None:
            return FunParams()

        # Must be deep-copied due to modification
        param_data = deepcopy(self.available_parameters[parameters_id])

        # Prepare the dictionary as input
        param_data = cast(FunParamsArgs, param_data)  # for type checker
        param_data["parameter_id"] = parameters_id

        # Deal with variable dimension parameter
        declared_parameters = param_data["declared_parameters"]
        for declared_parameter in declared_parameters:
            value = declared_parameter["value"]
            if isinstance(value, Callable):  # type: ignore
                func_signature = signature(value).parameters
                if "input_dimension" in func_signature:
                    parameter_value = value(input_dimension=input_dim)
                    declared_parameter["value"] = parameter_value

        return FunParams(**param_data)


def _init_subclass(cls: Type[UQTestFunABC]):
    """Verify if concrete class has all the required hidden attributes.

    Raises
    ------
    NotImplementedError
        If required attributes are not implemented in the concrete class.
    ValueError
        If default input and parameters selections are not specified
        when there are multiple of them.
    KeyError
        If the selections for the default input and parameters set are
        not available.
    """
    for class_hidden_attribute in CLASS_HIDDEN_ATTRIBUTES:
        # Some class attributes must be specified
        if not hasattr(cls, class_hidden_attribute):
            raise NotImplementedError(
                f"Class {cls} lacks required {class_hidden_attribute!r} "
                f"class attribute."
            )

    # Parse default input selection
    if cls.default_input_id:
        if cls.default_input_id not in cls.available_inputs:
            raise KeyError("Input selection is not available!")
    else:
        if len(cls.available_inputs) > 1:
            raise ValueError(
                "There are multiple available input specifications, "
                "the default input selection must be specified!"
            )
        else:
            # If only one is available, use it without being specified
            cls._default_input_id = list(cls.available_inputs.keys())[0]

    # Parse default parameters set selection
    if cls.available_parameters:
        if cls.default_parameters_id:
            if cls.default_parameters_id not in cls.available_parameters:
                raise KeyError("Parameters selection is not available!")

        else:
            if len(cls.available_parameters) > 1:
                raise ValueError(
                    "There are multiple available parameters sets, "
                    "the default input selection must be specified!"
                )
            else:
                # If only one is available, use it without being specified
                cls._default_parameters_id = list(
                    cls.available_parameters.keys()
                )[0]


def _verify_sample_shape(xx: np.ndarray, num_cols: int):
    """Verify the number of columns of the input sample array.

    Parameters
    ----------
    xx : np.ndarray
        Array of sampled input values with a shape of N-by-M, where N is
        the number of realizations and M is the input dimension.
    num_cols : int
        The expected number of columns in the input.

    Raises
    ------
    ValueError
        If the number of columns in the input is not equal to the expected
        number of columns.
    """
    if xx.shape[1] != num_cols:
        raise ValueError(
            f"Wrong dimensionality of the input array!"
            f"Expected {num_cols}, got {xx.shape[1]}."
        )


def _verify_sample_domain(xx: np.ndarray, min_value: float, max_value: float):
    """Verify whether the sampled input values are within the min and max.

    Parameters
    ----------
    xx : np.ndarray
        Array of sampled input values with a shape of N-by-M, where N is
        the number of realizations and M is the input dimension.
    min_value : float
        The minimum value of the domain.
    max_value : float
        The maximum value of the domain.

    Raises
    ------
    ValueError
        If any of the input values are outside the domain.
    """
    if not np.all(np.logical_and(xx >= min_value, xx <= max_value)):
        raise ValueError(
            f"One or more values are outside the domain "
            f"[{min_value}, {max_value}]!"
        )



"""
Module with an implementation of the simple 3D portfolio model.

The simple portfolio model is a three-dimensional scalar-valued function.
The function was first introduced in [1] as an example for illustrating
different sensitivity measures (i.e., local and hybrid-local).

References
----------

1. A. Saltelli, S. Tarantola, F. Campolongo, M. Rattom, "Sensitivity analysis
   in practice: a guide to assessing scientific models," Hoboken, NJ: Wiley,
   2004.
"""

import numpy as np

from uqtestfuns.core.custom_typing import (
    MarginalSpecs,
    ProbInputSpecs,
    FunParamSpecs,
)
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC


__all__ = ["Portfolio3D"]


MARGINALS_SALTELLI2004: MarginalSpecs = [
    {
        "name": "Ps",
        "distribution": "normal",
        "parameters": [0.0, 4.0],
        "description": "Hedged portfolio 's' [\N{euro sign}]",
    },
    {
        "name": "Pt",
        "distribution": "normal",
        "parameters": [0.0, 2.0],
        "description": "Hedged portfolio 't' [\N{euro sign}]",
    },
    {
        "name": "Pj",
        "distribution": "normal",
        "parameters": [0.0, 1.0],
        "description": "Hedged portfolio 'j' [\N{euro sign}]",
    },
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Saltelli2004": {
        "function_id": "Portfolio3D",
        "description": (
            "Probabilistic input model for the simple portfolio model "
            "from Saltelli et al. (2004)."
        ),
        "marginals": MARGINALS_SALTELLI2004,
        "copulas": None,
    },
}

AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Saltelli2004-1": {
        "function_id": "Portfolio3D",
        "description": (
            "Parameter set for the simple 3D portfolio model from "
            "Saltelli et al. (2004); the least volatile hedged portfolio is "
            "the largest quantity"
        ),
        "declared_parameters": [
            {
                "keyword": "cs",
                "value": 100.0,
                "type": float,
                "description": "Quantities of the portfolio 's'",
            },
            {
                "keyword": "ct",
                "value": 500.0,
                "type": float,
                "description": "Quantities of the portfolio 't'",
            },
            {
                "keyword": "cj",
                "value": 1000.0,
                "type": float,
                "description": "Quantities of the portfolio 'j'",
            },
        ],
    },
    "Saltelli2004-2": {
        "function_id": "Portfolio3D",
        "description": (
            "Parameter set for the simple 3D portfolio model from "
            "Saltelli et al. (2004); the same quantities are hold for each "
            "hedged portfolio"
        ),
        "declared_parameters": [
            {
                "keyword": "cs",
                "value": 300.0,
                "type": float,
                "description": "Quantities of the portfolio 's'",
            },
            {
                "keyword": "ct",
                "value": 300.0,
                "type": float,
                "description": "Quantities of the portfolio 't'",
            },
            {
                "keyword": "cj",
                "value": 300.0,
                "type": float,
                "description": "Quantities of the portfolio 'j'",
            },
        ],
    },
    "Saltelli2004-3": {
        "function_id": "Portfolio3D",
        "description": (
            "Parameter set for the simple 3D portfolio model from "
            "Saltelli et al. (2004); the most volatile hedged portfolio is"
            "the largest quantity"
        ),
        "declared_parameters": [
            {
                "keyword": "cs",
                "value": 500.0,
                "type": float,
                "description": "Quantities of the portfolio 's'",
            },
            {
                "keyword": "ct",
                "value": 400.0,
                "type": float,
                "description": "Quantities of the portfolio 't'",
            },
            {
                "keyword": "cj",
                "value": 100.0,
                "type": float,
                "description": "Quantities of the portfolio 'j'",
            },
        ],
    },
}

DEFAULT_PARAMETERS_SELECTION = "Saltelli2004-1"


def evaluate(xx: np.ndarray, cs: float, ct: float, cj: float) -> np.ndarray:
    """Evaluate the simple portfolio model on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Three-Dimensional input values given by an ``(N, 3)`` array where
        ``N`` is the number of input values.
    cs : float
        The quantities of the portfolio 's' (the most volatile).
    ct : float
        The quantities of the portfolio 't' (average volatility).
    cj : float
        The quantities of the portfolio 'j' (the least volatile).

    Returns
    -------
    np.ndarray
        The output of the simple portfolio model on the given input values.
        The output is a one-dimensional array of length ``N``.
    """
    # Read the parameters
    p = np.array([cs, ct, cj])

    # Compute the simple portfolio model
    yy = np.sum(p * xx, axis=1)

    return yy


class Portfolio3D(UQTestFunFixDimABC):
    """An implementation of the simple portfolio model test function."""

    _tags = ["sensitivity"]
    _description = "Simple portfolio model from Saltelli et al. (2004)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS
    _default_parameters_id = DEFAULT_PARAMETERS_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
The init for the 'test_functions' subpackage of UQTestFuns.
"""

from .ackley import Ackley
from .alemazkoor import Alemazkoor2D, Alemazkoor20D
from .borehole import Borehole
from .bratley1992 import Bratley1992a, Bratley1992b, Bratley1992c, Bratley1992d
from .cantilever_beam_2d import CantileverBeam2D
from .cheng2010 import Cheng2D
from .circular_pipe_crack import CircularPipeCrack
from .convex_fail_domain import ConvexFailDomain
from .damped_cosine import DampedCosine
from .damped_oscillator import DampedOscillator, DampedOscillatorReliability
from .flood import Flood
from .forrester import Forrester2008
from .four_branch import FourBranch
from .franke import Franke1, Franke2, Franke3, Franke4, Franke5, Franke6
from .friedman import Friedman6D, Friedman10D
from .gayton_hat import GaytonHat
from .gramacy2007 import Gramacy1DSine
from .hyper_sphere import HyperSphere
from .ishigami import Ishigami
from .lim import LimPoly, LimNonPoly
from .oakley2002 import Oakley1D
from .otl_circuit import OTLCircuit
from .mclain import McLainS1, McLainS2, McLainS3, McLainS4, McLainS5
from .moon3d import Moon3D
from .morris2006 import Morris2006
from .piston import Piston
from .portfolio_3d import Portfolio3D
from .rs_circular_bar import RSCircularBar
from .rs_quadratic import RSQuadratic
from .saltelli_linear import SaltelliLinear
from .sobol_g import SobolG
from .sobol_g_star import SobolGStar
from .sobol_levitan import SobolLevitan
from .solar_cell import SolarCell
from .speed_reducer_shaft import SpeedReducerShaft
from .sulfur import Sulfur
from .undamped_oscillator import UndampedOscillator
from .webster import Webster2D
from .welch1992 import Welch1992
from .wing_weight import WingWeight

# NOTE: Import the new test function implementation class from its respective
# module manually here and update the list below.

__all__ = [
    "Ackley",
    "Alemazkoor2D",
    "Alemazkoor20D",
    "Borehole",
    "Bratley1992a",
    "Bratley1992b",
    "Bratley1992c",
    "Bratley1992d",
    "CantileverBeam2D",
    "Cheng2D",
    "CircularPipeCrack",
    "ConvexFailDomain",
    "DampedCosine",
    "DampedOscillator",
    "DampedOscillatorReliability",
    "Flood",
    "Forrester2008",
    "FourBranch",
    "Franke1",
    "Franke2",
    "Franke3",
    "Franke4",
    "Franke5",
    "Franke6",
    "Friedman6D",
    "Friedman10D",
    "GaytonHat",
    "Gramacy1DSine",
    "HyperSphere",
    "Ishigami",
    "LimNonPoly",
    "LimPoly",
    "Oakley1D",
    "OTLCircuit",
    "McLainS1",
    "McLainS2",
    "McLainS3",
    "McLainS4",
    "McLainS5",
    "Moon3D",
    "Morris2006",
    "Piston",
    "Portfolio3D",
    "RSCircularBar",
    "RSQuadratic",
    "SaltelliLinear",
    "SobolG",
    "SobolGStar",
    "SobolLevitan",
    "SolarCell",
    "SpeedReducerShaft",
    "Sulfur",
    "UndampedOscillator",
    "Webster2D",
    "Welch1992",
    "WingWeight",
]



"""
Module with routines involving the Triangular probability distribution.

The Triangular distribution in UQTestFuns is parameterized
by three parameters: a, b, and c.
These are the lower bound, upper bound, and the mid-point (i.e., modal value),
respectively.

The density value at the modal value is computed such that the are under
the distribution is 1.0.
"""

import numpy as np

from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "triangular"

NUM_PARAMS = 3


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a triangular distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a triangular distribution (i.e., the lower bound,
        upper bound, and modal value).

    Returns
    -------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shape is inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    if not (parameters[0] < parameters[2] < parameters[1]):
        raise ValueError(
            f"The parameters must be: lower bound {parameters[0]} "
            f"< upper bound {parameters[1]} "
            f"< modal value {parameters[2]}."
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a triangular distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a triangular distribution.

    Returns
    -------
    float
        The lower bound of the triangular distribution.
    """
    lower_bound = float(parameters[0])

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a triangular distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a triangular distribution.

    Returns
    -------
    float
        The upper bound of a triangular distribution.
    """
    upper_bound = float(parameters[1])

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a triangular distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a triangular distribution.
    parameters : ARRAY_FLOAT
        Parameters of the triangular distribution.
    lower_bound : float
        Lower bound of the triangular distribution.
    upper_bound : float
        Upper bound of the triangular distribution.

    Returns
    -------
    ARRAY_FLOAT
        The PDF values of the triangular distribution.

    Notes
    -----
    - The values outside the bounds are set to 0.0.
    """
    mid_point = parameters[2]
    yy = np.zeros(xx.shape)
    idx_below_mid = np.logical_and(xx >= lower_bound, xx <= mid_point)
    idx_above_mid = np.logical_and(xx > mid_point, xx <= upper_bound)

    modal_density = 2 / (upper_bound - lower_bound)
    yy[idx_below_mid] = (
        modal_density
        / (mid_point - lower_bound)
        * (xx[idx_below_mid] - lower_bound)
    )
    yy[idx_above_mid] = (
        modal_density
        / (upper_bound - mid_point)
        * (upper_bound - xx[idx_above_mid])
    )

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of a triangular distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a triangular distribution.
    parameters : ARRAY_FLOAT
        Parameters of the triangular distribution.
    lower_bound : float
        Lower bound of the triangular distribution.
    upper_bound : float
        Upper bound of the triangular distribution.

    Returns
    -------
    ARRAY_FLOAT
        The CDF values of the triangular distribution.

    Notes
    -----
    - The CDF values for sample values below the lower bound are set to 0.0,
      and for sample values above the upper bound are set to 1.0.
    """

    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    mid_point = parameters[2]
    idx_below_mid = np.logical_and(xx >= lower_bound, xx <= mid_point)
    idx_above_mid = np.logical_and(xx > mid_point, xx <= upper_bound)

    yy = np.empty(xx.shape)
    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_below_mid] = (
        (xx[idx_below_mid] - lower_bound) ** 2
        / (upper_bound - lower_bound)
        / (mid_point - lower_bound)
    )
    yy[idx_above_mid] = 1 - (upper_bound - xx[idx_above_mid]) ** 2 / (
        upper_bound - lower_bound
    ) / (upper_bound - mid_point)

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a triangular distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT:
        Sample values (realizations) in the [0, 1] domain.
    parameters : ARRAY_FLOAT
        Parameters of the triangular distribution.
    lower_bound : float
        Lower bound of the triangular distribution.
    upper_bound : float
        Upper bound of the triangular distribution.

    Returns
    -------
    ARRAY_FLOAT
        Transformed values in the domain of the triangular distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN.
    """
    mid_point = parameters[2]
    mid_point_quantile = (mid_point - lower_bound) / (
        (upper_bound - lower_bound)
    )
    idx_below_mid = np.logical_and(xx >= 0.0, xx <= mid_point_quantile)
    idx_above_mid = np.logical_and(xx > mid_point_quantile, xx <= 1.0)
    idx_nan = np.logical_not(np.logical_or(idx_below_mid, idx_above_mid))

    yy = np.empty(xx.shape)
    yy[idx_below_mid] = lower_bound + np.sqrt(
        xx[idx_below_mid]
        * (upper_bound - lower_bound)
        * (mid_point - lower_bound)
    )

    yy[idx_above_mid] = upper_bound - np.sqrt(
        (1 - xx[idx_above_mid])
        * (upper_bound - lower_bound)
        * (upper_bound - mid_point)
    )

    yy[idx_nan] = np.nan

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Module with routines involving the Beta probability distribution.

The Beta distribution in UQTestFuns is parameterized by four parameters:
``r``, ``s``, ``a``, and ``b``. ``r`` and ``s`` are the shape parameters,
while ``a`` and ``b`` are the lower and upper bounds, respectively.

The underlying implementation is based on the implementation of scipy.stats.
In the SciPy implementation, the Beta distribution is parameterized also
by four parameters: ``a``, ``b``, ``loc``, and ``scale``.
``a`` and ``b`` are the shape parameters, while ``loc`` and ``scale`` are
the shifting and scaling parameters, respectively.

The translation between parameterization of the distribution in UQTestFuns
and SciPy is as follows:

- ``a`` == ``r``
- ``b`` == ``s``
- ``loc`` == ``a``
- ``scale`` == ``(b-a)``
"""

import numpy as np

from scipy.stats import beta

from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT


DISTRIBUTION_NAME = "beta"

NUM_PARAMS = 4


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a Beta distribution.

    Parameters
    ----------
    parameters : np.ndarray
        The parameters of a Beta distribution
        (i.e., the two shape parameters and the lower and upper bounds).

    Returns
    ------
    None
        Function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Check overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    # Check validity of values
    if parameters[0] <= 0.0 or parameters[1] <= 0.0:
        raise ValueError(
            f"The shape parameters {parameters[0]} and {parameters[1]}"
            f"must be larger than 0.0 (positive)!"
        )

    if parameters[2] >= parameters[3]:
        raise ValueError(
            f"The lower bound {parameters[2]} "
            f"cannot be equal or greater than the upper bound {parameters[3]}!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a Beta distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a Beta distribution.

    Returns
    -------
    float
        The lower bound of the Beta distribution.
    """
    lower_bound = float(parameters[2])

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a Beta distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a Beta distribution

    Returns
    -------
    float
        The upper bound of the Beta distribution.
    """
    upper_bound = float(parameters[3])

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a Beta distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a Beta distribution.
    parameters : ARRAY_FLOAT
        Parameters of the Beta distribution.
    lower_bound : float
        Lower bound of the Beta distribution.
    upper_bound : float
        Upper bound of the Beta distribution.

    Returns
    -------
    ARRAY_FLOAT
        PDF values of the Beta distribution on the sample values.

    Notes
    -----
    - The values outside the bounds are set to 0.0.
    """
    yy = np.zeros(xx.shape)
    idx = np.logical_and(xx >= lower_bound, xx <= upper_bound)
    yy[idx] = beta.pdf(
        xx[idx],
        a=parameters[0],
        b=parameters[1],
        loc=parameters[2],
        scale=parameters[3] - parameters[2],
    )

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of the Beta distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a Beta distribution.
    parameters : ARRAY_FLOAT
        Parameters of the Beta distribution.
    lower_bound : float
        Lower bound of the Beta distribution.
    upper_bound : float
        Upper bound of the Beta distribution.

    Returns
    -------
    ARRAY_FLOAT
        CDF values of the Beta distribution on the sample values.

    Notes
    -----
    - CDF for sample with values smaller (resp. larger) than the lower bound
      (resp. upper bound) are set to 0.0 (resp. 1.0).
    """
    yy = np.empty(xx.shape)
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_rest] = beta.cdf(
        xx[idx_rest],
        a=parameters[0],
        b=parameters[1],
        loc=parameters[2],
        scale=parameters[3] - parameters[2],
    )

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a Beta distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) in the [0, 1] domain.
    parameters : ARRAY_FLOAT
        Parameters of a Beta distribution.
    lower_bound : float
        Lower bound of the Beta distribution.
    upper_bound : float
        Upper bound of the Beta distribution.

    Returns
    -------
    ARRAY_FLOAT
        Transformed values in the domain of the Beta distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN according to
      the underlying SciPy implementation.
    """
    # Get the parameters
    r, s, a, b = parameters[:]

    # Compute the ICDF
    yy = beta.ppf(xx, a=r, b=s, loc=a, scale=b - a)

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
This module contains the implementation of FunParams a class that stores
UQ Test function parameters.
"""

import textwrap

from tabulate import tabulate
from typing import Any, List, Optional

import warnings

__all__ = ["FunParams"]

import numpy as np

from uqtestfuns.core.custom_typing import DeclaredParameters

FIELD_NAMES = ["Keyword", "Value", "Type", "Description"]


class FunParams:
    """A class that stores UQ Test function parameters.

    Parameters
    ----------
    function_id : str, optional.
        The UQ test function ID (typically the name of the test function).
        If not provided, an empty string is used.
    parameter_id : str, optional
        The ID of the parameter set.
        If not provided, an empty string is used.
    description : str, optional
        Descriptive string of the parameter set.
        If not provided, an empty string is used.
    declared_parameters : List[dict], optional
        List of dictionaries containing the UQ test function parameters with
        the following keywords: 'keyword', 'value', 'type', 'description'.
        'keyword' value is a string that appears in the evaluation function.
        The value is None if not provided.
    """

    def __init__(
        self,
        function_id: str = "",
        parameter_id: str = "",
        description: str = "",
        declared_parameters: Optional[DeclaredParameters] = None,
    ):
        self._declared_parameters: dict = {}
        self._dict: dict = {}

        self.function_id = function_id
        self.parameter_id = parameter_id
        self.description = description

        if declared_parameters is None:
            return

        # Parse the declared parameters
        for parameter in declared_parameters:
            self.add(**parameter)

    def __getitem__(self, key: str):
        """Get the current value by name.

        Parameters
        ----------
        key : str
            The name of the parameter, i.e., its keyword. The parameter
            must already be declared.
        """
        return self._dict[key]

    def __setitem__(self, key: str, value: Any):
        """Set the current value by name.

        Parameters
        ----------
        key : str
            The name of the parameter, i.e., its keyword. The parameter must
            already be declared.
        value : Any
            The value of the parameter.

        Notes
        -----
        - This method allows the parameter value to be changed after
          creation.
        """
        if key not in self._dict:
            raise KeyError(f"Parameter '{key}' is not declared")

        self._assert_value_assign(key, value)

        self._dict[key] = value

    def __len__(self) -> int:
        """Return the length of the parameter set."""
        return len(self._declared_parameters)

    def __bool__(self) -> bool:
        """Return True if one or more parameters are declared."""
        return self.__len__() > 0

    def __eq__(self, other) -> bool:
        """Check the equality in value between two instances.

        Parameters
        ----------
        other
            The right-hand side of the operation.

        Notes
        -----
        - Only the equality in the constituent parameters is checked.
        """
        if not isinstance(other, FunParams):
            return False

        a = _nested_dicts_equal(self._dict, other._dict)
        b = _nested_dicts_equal(
            self._declared_parameters,
            other._declared_parameters,
        )

        return a and b

    def __str__(self):
        if not self:
            return str(None)

        table = f"Function ID  : {self.function_id}\n"
        table += f"Parameter ID : {self.parameter_id}\n"
        desc = textwrap.wrap(self.description, width=57)
        # Pad new lines
        if len(desc) > 1:
            desc[1:] = ["               " + line for line in desc[1:]]
        desc_joined = "\n".join(desc)
        table += f"Description  : {desc_joined}\n\n"

        # Get the header names
        header_names = [name.capitalize() for name in FIELD_NAMES]
        header_names.insert(0, "No.")

        # Get the values for each field as a list
        list_values = _get_values_as_list(
            self._dict,
            self._declared_parameters,
        )

        table += tabulate(
            list_values,
            headers=header_names,
            stralign="center",
            disable_numparse=True,
        )

        return table

    def as_dict(self):
        """Return key-value pairs of the parameter set."""
        return self._dict

    def add(
        self,
        keyword: str,
        value: Any,
        type: Optional[type] = None,
        description: Optional[str] = "",
    ) -> None:
        """Add a parameter to the parameter set.

        Parameters
        ----------
        keyword : str
            The name of the parameter as it appears in the evaluation function
            signature; must be a valid Python identifier.
        value : Any
            The default value of the parameter.
        type : Optional[type]
            The type of the parameter. If given, the type will be used to
            validate the provided value.
        description : str, optional
            The description of the parameter.
        """
        if keyword in self._declared_parameters:
            raise KeyError(f"Parameter '{keyword}' is already defined")

        if not keyword.isidentifier():
            raise ValueError(
                f"Parameter '{keyword}' is not a valid identifier"
            )

        self._declared_parameters[keyword] = {
            "value": value,
            "type": type,
            "description": description,
        }

        # Verify the value here
        self._assert_value_set(keyword)

        # Update the underlying dict
        self._dict[keyword] = value

    def reset_value(self) -> None:
        """Reset the value of the parameter set to the declared values."""
        self._dict.clear()
        for key, value in self._declared_parameters.items():
            self._dict[key] = value["value"]

    def _assert_value_set(self, keyword: str):
        """Assert the parameter set when they're added the first time.

        Parameters
        ----------
        keyword : str
            The name of the parameter.

        Notes
        -----
        - The type of the value must be consistent with the type (if given).
        """
        type_ = self._declared_parameters[keyword]["type"]
        value = self._declared_parameters[keyword]["value"]
        if type_ is None:
            return

        # When a parameter is first added enforce consistency
        if not isinstance(value, type_):
            raise TypeError(f"Expected {type_} but got {type(value)}")

    def _assert_value_assign(self, keyword, value):
        """Assert the parameter set when they're assigned after creation.

        Parameters
        ----------
        keyword : str
            The name of the parameter.
        value : Any
            The value of the parameter.
        Notes
        -----
        - If the value and the type are not consistent, a warning (but no
          exception is raised).
        """
        type_ = self._declared_parameters[keyword]["type"]
        if type_ is None:
            return

        # When a parameter is assigned, be more forgiving
        if not isinstance(value, type_):
            warnings.warn(
                message="Expected {type_} but got {type(value)}",
                category=UserWarning,
                stacklevel=2,
            )


def _nested_dicts_equal(d1, d2) -> bool:
    """Compare the equality in value between two nested dictionaries.

    Parameters
    ----------
    d1
        The first dictionary.
    d2
        The second dictionary

    Returns
    -------
    bool
        True if the dictionaries are equal, False otherwise.
    """
    if not isinstance(d1, dict) or not isinstance(d2, dict):

        # Deal specifically with NumPy array values
        if isinstance(d1, np.ndarray) and isinstance(d2, np.ndarray):
            return np.array_equal(d1, d2)
        return d1 == d2  # Compare values directly if not dictionaries

    if d1.keys() != d2.keys():
        return False  # Different keys

    # Recursively check each key
    return all(_nested_dicts_equal(d1[key], d2[key]) for key in d1)


def _get_values_as_list(
    current_values: dict,
    declared_parameters: dict,
) -> List[List[str]]:
    """

    Parameters
    ----------
    current_values : dict
        The dictionary of current values.
    declared_parameters : dict
        The dictionaries of declared parameters.

    Return
    ------
    List[List[str]]
        List of lists of the relevant data as strings.
    """
    list_values = []
    for i, parameter in enumerate(declared_parameters):
        # Value
        value = current_values[parameter]
        if isinstance(value, np.ndarray):
            value = f"{value.shape} array"
        elif isinstance(value, int):
            value = f"{value:1.0e}"
        elif isinstance(value, float):
            value = f"{value:1.5e}"
        else:
            value = str(value)

        # Type
        type_ = declared_parameters[parameter]["type"]
        if type_ is not None:
            type_ = type_.__name__
        else:
            type_ = "-"

        # Description
        desc = declared_parameters[parameter]["description"]
        if desc == "":
            desc = "-"

        entry = [
            i + 1,
            parameter,
            value,
            type_,
            desc,
        ]

        list_values.append(entry)

    return list_values



"""
Module with an implementation of the Welch et al. (1992) test function.

The Welch1992 test function is a 20-dimensional scalar-valued function.
The function features some strong non-linear effects as well as some pair
interaction effects. Furthermore, two input variables (namely x8 and x16)
are inert.

The function was introduced in [1] as a test function for metamodeling and
sensitivity analysis purposes. The function is also suitable for testing
multi-dimensional integration algorithms.

References
----------

1. William J. Welch, Robert J. Buck, Jerome Sacks, Henry P. Wynn,
   Toby J. Mitchell, and Max D. Morris, "Screening, predicting, and computer
   experiments," Technometrics, vol. 34, no. 1, pp. 15-25, 1992.
   DOI: 10.2307/1269548
"""

import numpy as np

from uqtestfuns.core.custom_typing import MarginalSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Welch1992"]


MARGINALS_WELCH1992: MarginalSpecs = [
    {
        "name": f"x{i}",
        "distribution": "uniform",
        "parameters": [-0.5, 0.5],
        "description": None,
    }
    for i in range(1, 20 + 1)
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Welch1992": {
        "function_id": "Welch1992",
        "description": (
            "Input specification for the test function "
            "from Welch et al. (1992)"
        ),
        "marginals": MARGINALS_WELCH1992,
        "copulas": None,
    },
}


DEFAULT_INPUT_SELECTION = "Welch1992"


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the Welch et al. (1992) function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        1-Dimensional input values given by an N-by-1 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.

    Notes
    -----
    - The input variables xx[:, 7] (x8) and xx[:, 15] (x16) are inert and
      therefore, does not appear in the computation below.
    """
    yy = (
        (5 * xx[:, 11]) / (1 + xx[:, 0])
        + 5 * (xx[:, 3] - xx[:, 19]) ** 2
        + xx[:, 4]
        + 40 * xx[:, 18] ** 3
        - 5 * xx[:, 18]
        + 0.05 * xx[:, 1]
        + 0.08 * xx[:, 2]
        - 0.03 * xx[:, 5]
        + 0.03 * xx[:, 6]
        - 0.09 * xx[:, 8]
        - 0.01 * xx[:, 9]
        - 0.07 * xx[:, 10]
        + 0.25 * xx[:, 12] ** 2
        - 0.04 * xx[:, 13]
        + 0.06 * xx[:, 14]
        - 0.01 * xx[:, 16]
        - 0.03 * xx[:, 17]
    )

    return yy


class Welch1992(UQTestFunFixDimABC):
    """A concrete implementation of the Welch et al. (1992) test function."""

    _tags = ["metamodeling", "sensitivity", "integration"]
    _description = "20-Dimensional function from Welch et al. (1992)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the Quadratic RS reliability problem.

The two-dimensional function is a variant of the classic RS reliability problem
with one quadratic term [1].

References
----------
1. Paul Hendrik Waarts, “Structural reliability using finite element
   analysis - an appraisal of DARS: Directional adaptive response surface
   sampling," Civil Engineering and Geosciences, TU Delft, Delft,
   The Netherlands, 2000.
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["RSQuadratic"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Waarts2000": {
        "function_id": "RSQuadratic",
        "description": "Input model for the quadratic RS from Waarts (2000)",
        "marginals": [
            {
                "name": "X1",
                "distribution": "normal",
                "parameters": [11.0, 1.0],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "normal",
                "parameters": [1.5, 0.5],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the quadratic RS function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by an N-by-2 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The performance function of the problem.
        If negative, the system is in failed state.
        The output is a one-dimensional array of length N.
    """

    # Compute the performance function
    yy = xx[:, 0] - xx[:, 1] ** 2

    return yy


class RSQuadratic(UQTestFunFixDimABC):
    """Concrete implementation of the quadratic RS reliability problem."""

    _tags = ["reliability"]
    _description = "RS problem w/ one quadratic term from Waarts (2000)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_dimension = 2

    evaluate = staticmethod(evaluate)  # type: ignore



"""
This module implements the Morris function from Morris et al. (2006).

The Morris function is an M-dimensional, scalar-valued function commonly used
as a test function for sensitivity analysis.
It was first appeared in [1] and it has been revisited multiple times in the
literature in similar contexts, e.g., [2], [3].

The function features a parameter that dictates the number of important
input variables. The remaining input variables are inert. Furthermore,
all sensitivity indices of the important variables have
the same Sobol' sensitivity indices.

References
----------

1. M. D. Morris, L. M. Moore, and M. D. McKay, “Sampling plans based on
   balanced incomplete block designs for evaluating the importance of
   computer model inputs,” Journal of Statistical Planning and Inference,
   vol. 136, no. 9, pp. 3203–3220, 2006.
   DOI: 10.1016/j.jspi.2005.01.001
2. X. Sun, B. Croke, A. Jakeman, S. Roberts, "Benchmarking Active Subspace
   methods of global sensitivity analysis against variance-based Sobol’
   and Morris methods with established test functions," Environmental Modelling
   & Software, vol. 149, p. 105310, 2022.
   DOI: 10.1016/j.envsoft.2022.105310
3. A. Horiguchi, M. T. Pratola, and T. J. Santner, “Assessing variable activity
   for Bayesian regression trees,” Reliability Engineering & System Safety,
   vol. 207, p. 107391, 2021,
   DOI: 10.1016/j.ress.2020.107391
"""

import numpy as np

from uqtestfuns.core.custom_typing import FunParamSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunVarDimABC

__all__ = ["Morris2006"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Morris2006": {
        "function_id": "Morris2006",
        "description": (
            "Probabilistic input model for the M-dimensional function "
            "from Morris et al. (2006)"
        ),
        "marginals": [
            {
                "name": "X",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Morris2006": {
        "function_id": "Morris2006",
        "description": (
            "Parameter set for the M-dimensional function from "
            "Morris et al. (2006); the parameter controls the number of "
            "important input variables"
        ),
        "declared_parameters": [
            {
                "keyword": "p",
                "value": 10,
                "type": int,
                "description": "# of important inputs",
            },
        ],
    }
}


def evaluate(xx: np.ndarray, p: int) -> np.ndarray:
    """Evaluate the Morris2006 test function.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.
    p : int
        The number of important input variables. If p > M then p is set
        equal to M.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    input_dim = xx.shape[1]
    if p > input_dim:
        p = input_dim

    alpha = np.sqrt(12) - 6 * np.sqrt(0.1 * (p - 1))

    term_1 = alpha * np.sum(xx[:, :p], axis=1)
    if p == 1:
        return term_1

    beta = 12 / np.sqrt(10 * (p - 1))
    term_2 = np.zeros(len(xx))
    for i in range(p - 1):
        ip1 = i + 1
        term_2[:] += xx[:, i] * np.sum(xx[:, ip1:p], axis=1)
    term_2 *= beta

    yy = term_1 + term_2

    return yy


class Morris2006(UQTestFunVarDimABC):
    """An implementation of the M-dimensional Morris2006."""

    _tags = ["sensitivity"]
    _description = "Test function from Morris et al. (2006)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS

    evaluate = staticmethod(evaluate)  # type: ignore



"""
This module implements the Sobol'-Levitan test function.

The Sobol'-Levitan function is an M-dimensional, scalar-valued function
commonly used as a benchmark for sensitivity analysis.
The function was introduced in [1] (as a six- and 20-dimensional functions)
and revisited in, for example, [2] (as a 20-dimensional function)
and [3] (as a seven- and 15-dimensional functions).

The Sobol' sensitivity indices of the test function can be derived analytically
as functions of the parameter.

References
----------

1. I. M. Sobol’ and Yu. L. Levitan, “On the use of variance reducing
   multipliers in Monte Carlo computations of a global sensitivity index,”
   Computer Physics Communications, vol. 117, no. 1–2, pp. 52–61, 1999.
   DOI: 10.1016/S0010-4655(98)00156-8
2. H. Moon, A. M. Dean, and T. J. Santner, “Two-Stage Sensitivity-Based Group
   Screening in Computer Experiments,” Technometrics*, vol. 54, no. 4,
   pp. 376–387, 2012.
   DOI: 10.1080/00401706.2012.725994
3. X. Sun, B. Croke, A. Jakeman, S. Roberts, "Benchmarking Active Subspace
   methods of global sensitivity analysis against variance-based Sobol’
   and Morris methods with established test functions," Environmental Modelling
   & Software, vol. 149, p. 105310, 2022.
   DOI: 10.1016/j.envsoft.2022.105310
"""

import numpy as np

from uqtestfuns.core.custom_typing import FunParamSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunVarDimABC

__all__ = ["SobolLevitan"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Sobol1999": {
        "function_id": "SobolLevitan",
        "description": (
            "Probabilistic input model for the Sobol'-Levitan function "
            "from Sobol' and Levitan (1999)"
        ),
        "marginals": [
            {
                "name": "X",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def _get_bb_sobol_1999_1(input_dimension: int) -> np.ndarray:
    """Construct the coefficients from Sobol' and Levitan (1999) 6D case.

    Notes
    -----
    - In [1], an input dimension of 6 was used. If the input dimension is less
      than 6, the parameters array is truncated; if the input dimension exceed
      10, the parameters are is extrapolated.
    """
    bb = np.array([1.5, 0.9, 0.9, 0.9, 0.9, 0.9])

    if input_dimension > 6:
        bb_ext = 0.9 * np.ones(input_dimension - len(bb))
        bb = np.append(bb, bb_ext)

    return bb[:input_dimension]


def _get_bb_sobol_1999_2(input_dimension: int) -> np.ndarray:
    """Construct the coefficients from Sobol' and Levitan (1999) 20D case.

    Notes
    -----
    - In [1], an input dimension of 20 was used. If the input dimension is less
      than 20, the parameters array is truncated; if the input dimension exceed
      20, the parameters are is extrapolated.
    """
    bb_1 = 0.6 * np.ones(10)
    bb_2 = 0.4 * np.ones(10)
    bb = np.append(bb_1, bb_2)

    if input_dimension > 20:
        bb_ext = 0.4 * np.ones(input_dimension - len(bb))
        bb = np.append(bb, bb_ext)

    return bb[:input_dimension]


def _get_bb_moon_2012_1(input_dimension: int) -> np.ndarray:
    """Construct the coefficients from Moon et al. (2012) base case.

    Notes
    -----
    - In [1], an input dimension of 20 was used. If the input dimension is less
      than 20, the parameters array is truncated; if the input dimension exceed
      20, the parameters are is extrapolated.
    """
    bb = np.array(
        [
            2.0000,
            1.9500,
            1.9000,
            1.8500,
            1.8000,
            1.7500,
            1.7000,
            1.6500,
            0.4228,
            0.3077,
            0.2169,
            0.1471,
            0.0951,
            0.0577,
            0.0323,
            0.0161,
            0.0068,
            0.0021,
            0.0004,
            0.0000,
        ]
    )

    if input_dimension > 20:
        bb_ext = np.zeros(input_dimension - len(bb))
        bb = np.append(bb, bb_ext)

    return bb[:input_dimension]


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Sobol1999-1": {
        "function_id": "SobolLevitan",
        "description": (
            "Parameter set for the M-dimensional function from "
            "Sobol' and Levitan (1999), 6D case"
        ),
        "declared_parameters": [
            {
                "keyword": "bb",
                "value": _get_bb_sobol_1999_1,
                "type": np.ndarray,
                "description": "Coefficients 'b'",
            },
            {
                "keyword": "c0",
                "value": 0.0,
                "type": float,
                "description": "Constant term",
            },
        ],
    },
    "Sobol1999-2": {
        "function_id": "SobolLevitan",
        "description": (
            "Parameter set for the M-dimensional function from "
            "Sobol' and Levitan (1999), 20D case"
        ),
        "declared_parameters": [
            {
                "keyword": "bb",
                "value": _get_bb_sobol_1999_2,
                "type": np.ndarray,
                "description": "Coefficients 'b'",
            },
            {
                "keyword": "c0",
                "value": 0.0,
                "type": float,
                "description": "Constant term",
            },
        ],
    },
    "Moon2012-1": {
        "function_id": "SobolLevitan",
        "description": (
            "Parameter set for the M-dimensional function from "
            "Moon et al. (2012), base case"
        ),
        "declared_parameters": [
            {
                "keyword": "bb",
                "value": _get_bb_moon_2012_1,
                "type": np.ndarray,
                "description": "Coefficients 'b'",
            },
            {
                "keyword": "c0",
                "value": 0.0,
                "type": float,
                "description": "Constant term",
            },
        ],
    },
}


DEFAULT_PARAMETERS_SELECTION = "Sobol1999-1"


def evaluate(xx: np.ndarray, bb: np.ndarray, c0: float) -> np.ndarray:
    """Evaluate the Sobol'-Levitan test function

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.
    bb : np.ndarray
        The coefficients of the function which control the importance of each
        input variable.
    c0 : float
        The constant term of the function.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    input_dim = xx.shape[1]
    ii = 1.0
    for i in range(input_dim):
        b = bb[i]
        # Safeguard against zero-value coefficient
        if b == 0:
            ii *= 1.0  # the limit as b approaches 0
        else:
            ii *= (np.exp(b) - 1) / b

    yy = np.exp(np.sum(bb * xx, axis=1)) - ii + c0

    return yy


class SobolLevitan(UQTestFunVarDimABC):
    """An implementation of the M-dimensional Sobol'-Levitan function."""

    _tags = ["sensitivity"]
    _description = "Test function from Sobol' and Levitan (1999)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS
    _default_parameters_id = DEFAULT_PARAMETERS_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the Modified Sobol-G test function.

The modified Sobol'-G function, often written as Sobol'-G* function,
is a scalar-valued function in M dimensions.
It was first introduced in [1] to serve as a test function for sensitivity
analysis methods.
This function, which introduces shift and curvature parameters,
is a modification of the original Sobol'-G test function [2].

The function has found application in the study of sensitivity analysis, e.g.,
[3].

There are several sets of parameters used in the literature.

References
----------

1. A. Saltelli, P. Annoni, I. Azzini, F. Campolongo, M. Ratto,
   and S. Tarantola, “Variance based sensitivity analysis of model output.
   Design and estimator for the total sensitivity index,”
   Computer Physics Communications, vol. 181, no. 2, pp. 259–270, 2010.
   DOI: 10.1016/j.cpc.2009.09.018
2. A. Saltelli and I. M. Sobol’, “About the use of rank transformation in
   sensitivity analysis of model output,” Reliability Engineering
   & System Safety, vol. 50, no. 3, pp. 225–239, 1995.
   DOI: 10.1016/0951-8320(95)00099-2.
3. X. Sun, B. Croke, A. Jakeman, S. Roberts, "Benchmarking Active Subspace
   methods of global sensitivity analysis against variance-based Sobol’
   and Morris methods with established test functions," Environmental Modelling
   & Software, vol. 149, p. 105310, 2022.
   DOI: 10.1016/j.envsoft.2022.105310
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs, FunParamSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunVarDimABC

__all__ = ["SobolGStar"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Saltelli2010": {
        "function_id": "SobolGStar",
        "description": (
            "Probabilistic input model for the Sobol'-G* function "
            "from Saltelli et al. (2010)"
        ),
        "marginals": [
            {
                "name": "X",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def _get_aa_saltelli_2010_a(input_dimension: int) -> np.ndarray:
    """Construct a coefficients array for the Sobol'-G* function from [1].

    Notes
    -----
    - This value is used in test case 1, 3, and 5 (see Table 5 in [1]).
    - In [1], an input dimension of 10 was used. If the input dimension is less
      than 10, the parameters array is truncated; if the input dimension exceed
      10, the parameters are is extrapolated.
    - With the selected parameters, the function is of low effective dimension.
    """
    aa = 9 * np.ones(input_dimension)
    aa[:2] = 0

    return aa


def _get_aa_saltelli_2010_b(input_dimension: int) -> np.ndarray:
    """Construct a coefficients array for the Sobol'-G* function from [1].

    Notes
    -----
    - This value is used in test case 2, 5, and 6 (see Table 5 in [1]).
    - In [1], an input dimension of 10 was used. If the input dimension is less
      than 10, the parameters array is truncated; if the input dimension exceed
      10, the parameters are is extrapolated.
    """
    aa = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.8, 1, 2, 3, 4])

    if input_dimension > 10:
        aa_ext = np.arange(5, input_dimension - 5)
        aa = np.append(aa, aa_ext)

    return aa[:input_dimension]


def _get_alpha_saltelli_2010_a(input_dimension: int) -> np.ndarray:
    """Construct an alpha array for the Sobol'-G* function from [1].

    Notes
    -----
    - This value is used in test case 1 and 2 (see Table 5 in [1]).
    - In [1], an input dimension of 10 was used. If the input dimension is less
      than 10, the parameters array is truncated; if the input dimension exceed
      10, the parameters are is extrapolated.
    """
    alpha = np.ones(input_dimension)

    return alpha


def _get_alpha_saltelli_2010_b(input_dimension: int) -> np.ndarray:
    """Construct an alpha array for the Sobol'-G* function from [1].

    Notes
    -----
    - This value is used in test case 3 and 4 (see Table 5 in [1]).
    - In [1], an input dimension of 10 was used. If the input dimension is less
      than 10, the parameters array is truncated; if the input dimension exceed
      10, the parameters are is extrapolated.
    """
    alpha = 0.5 * np.ones(input_dimension)

    return alpha


def _get_alpha_saltelli_2010_c(input_dimension: int) -> np.ndarray:
    """Construct an alpha array for the Sobol'-G* function from [1].

    Notes
    -----
    - This value is used in test case 5 and 6 (see Table 5 in [1]).
    - In [1], an input dimension of 10 was used. If the input dimension is less
      than 10, the parameters array is truncated; if the input dimension exceed
      10, the parameters are is extrapolated.
    """
    alpha = 2 * np.ones(input_dimension)

    return alpha


def _get_delta_saltelli_2010(input_dimension: int) -> np.ndarray:
    """Construct a delta array for the Sobol'-G* function from [1].

    The parameter delta (shift parameter) is randomly generated
    from a uniform distribution in [0, 1].
    """
    rng = np.random.default_rng()
    delta = rng.random(input_dimension)

    return delta


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Saltelli2010-1": {
        "function_id": "SobolGStar",
        "description": (
            "Parameter set for the Sobol-G* function from Saltelli et al."
            "(2010), test case 1; low-effective dimension, easier than "
            "test case 2"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_aa_saltelli_2010_a,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
            {
                "keyword": "delta",
                "value": _get_delta_saltelli_2010,
                "type": np.ndarray,
                "description": "Shift parameter",
            },
            {
                "keyword": "alpha",
                "value": _get_alpha_saltelli_2010_a,
                "type": np.ndarray,
                "description": "Curvature parameter",
            },
        ],
    },
    "Saltelli2010-2": {
        "function_id": "SobolGStar",
        "description": (
            "Parameter set for the Sobol-G* function from Saltelli et al. "
            "(2010), test case 2; high-effective dimension, more difficult "
            "than test case 1"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_aa_saltelli_2010_b,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
            {
                "keyword": "delta",
                "value": _get_delta_saltelli_2010,
                "type": np.ndarray,
                "description": "Shift parameter",
            },
            {
                "keyword": "alpha",
                "value": _get_alpha_saltelli_2010_a,
                "type": np.ndarray,
                "description": "Curvature parameter",
            },
        ],
    },
    "Saltelli2010-3": {
        "function_id": "SobolGStar",
        "description": (
            "Parameter set for the Sobol-G* function from Saltelli et al. "
            "(2010), test case 3; concave version of test case 1"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_aa_saltelli_2010_a,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
            {
                "keyword": "delta",
                "value": _get_delta_saltelli_2010,
                "type": np.ndarray,
                "description": "Shift parameter",
            },
            {
                "keyword": "alpha",
                "value": _get_alpha_saltelli_2010_b,
                "type": np.ndarray,
                "description": "Curvature parameter",
            },
        ],
    },
    "Saltelli2010-4": {
        "function_id": "SobolGStar",
        "description": (
            "Parameter set for the Sobol-G* function from Saltelli et al. "
            "(2010), test case 4; concave version of test case 2"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_aa_saltelli_2010_b,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
            {
                "keyword": "delta",
                "value": _get_delta_saltelli_2010,
                "type": np.ndarray,
                "description": "Shift parameter",
            },
            {
                "keyword": "alpha",
                "value": _get_alpha_saltelli_2010_b,
                "type": np.ndarray,
                "description": "Curvature parameter",
            },
        ],
    },
    "Saltelli2010-5": {
        "function_id": "SobolGStar",
        "description": (
            "Parameter set for the Sobol-G* function from Saltelli et al. "
            "(2010), test case 5; convex version of test case 1"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_aa_saltelli_2010_a,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
            {
                "keyword": "delta",
                "value": _get_delta_saltelli_2010,
                "type": np.ndarray,
                "description": "Shift parameter",
            },
            {
                "keyword": "alpha",
                "value": _get_alpha_saltelli_2010_c,
                "type": np.ndarray,
                "description": "Curvature parameter",
            },
        ],
    },
    "Saltelli2010-6": {
        "function_id": "SobolGStar",
        "description": (
            "Parameter set for the Sobol-G* function from Saltelli et al. "
            "(2010), test case 6; convex version of test case 2"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_aa_saltelli_2010_b,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
            {
                "keyword": "delta",
                "value": _get_delta_saltelli_2010,
                "type": np.ndarray,
                "description": "Shift parameter",
            },
            {
                "keyword": "alpha",
                "value": _get_alpha_saltelli_2010_c,
                "type": np.ndarray,
                "description": "Curvature parameter",
            },
        ],
    },
}

DEFAULT_PARAMETERS_SELECTION = "Saltelli2010-1"


def evaluate(
    xx: np.ndarray,
    aa: np.ndarray,
    delta: np.ndarray,
    alpha: np.ndarray,
) -> np.ndarray:
    """Evaluate the modified Sobol-G function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.
    aa : np.ndarray
        The vector of parameters (i.e., coefficients) of the modified Sobol'-G
        function; the length of the vector is the same as the number
        of input dimensions.
    delta: np.ndarray
        The vector of shift parameters of length M.
    alpha: np.ndarray
        The vector of curvature parameters of length M.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    term = np.abs(2 * (xx + delta - np.floor(xx + delta)) - 1) ** alpha
    g_star = ((1 + alpha) * term + aa) / (1 + aa)

    yy = np.prod(g_star, axis=1)

    return yy


class SobolGStar(UQTestFunVarDimABC):
    """An implementation of the M-dimensional Sobol'-G* test function."""

    _tags = ["sensitivity"]
    _description = "Sobol'-G* function from Saltelli et al. (2010)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS
    _default_parameters_id = DEFAULT_PARAMETERS_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of Alemazkoor & Meidani (2008) test functions.

There are two test functions from [1].
One features a low-dimensional polynomial function with a high degree
(a total degree of 20).

The functions were used as test functions
for a metamodeling exercise (i.e., sparse polynomial chaos expansion) in [1].

References
----------

1. Negin Alemazkoor and Hadi Meidani, "A near-optimal sampling strategy for
   sparse recovery of polynomial chaos expansions," Journal of Computational
   Physics, vol. 371, pp. 137-151, 2018.
   DOI: 10.1016/j.jcp.2018.05.025
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Alemazkoor2D", "Alemazkoor20D"]


AVAILABLE_INPUTS_2D: ProbInputSpecs = {
    "Alemazkoor2018": {
        "function_id": "Alemazkoor2D",
        "description": (
            "Input specification for the 2D test function "
            "from Alemazkoor & Meidani (2018)"
        ),
        "marginals": [
            {
                "name": "X1",
                "distribution": "uniform",
                "parameters": [-1.0, 1.0],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "uniform",
                "parameters": [-1.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}

AVAILABLE_INPUTS_20D: ProbInputSpecs = {
    "Alemazkoor2018": {
        "function_id": "Alemazkoor20D",
        "description": (
            "Input specification for the 20D test function "
            "from Alemazkoor & Meidani (2018)"
        ),
        "marginals": [
            {
                "name": f"X{i + 1}",
                "distribution": "uniform",
                "parameters": [-1, 1],
                "description": None,
            }
            for i in range(20)
        ],
        "copulas": None,
    },
}


def evaluate_2d(xx: np.ndarray) -> np.ndarray:
    """Evaluate the 2D test function from Alemazkoor & Meidani (2018).

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by an ``(N, 2)`` array where
        ``N`` is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a one-dimensional array of length ``N``.
    """
    yy = np.zeros(xx.shape[0])

    for i in range(1, 6):
        yy += xx[:, 0] ** (2 * i) * xx[:, 1] ** (2 * i)

    return yy


class Alemazkoor2D(UQTestFunFixDimABC):
    """An implementation of the 2D function of Alemazkoor & Meidani (2018)."""

    _tags = ["metamodeling"]
    _description = (
        "Low-dimensional high-degree polynomial from Alemazkoor "
        "& Meidani (2018)"
    )
    _available_inputs = AVAILABLE_INPUTS_2D
    _available_parameters = None
    _default_input_dimension = 2

    evaluate = staticmethod(evaluate_2d)  # type: ignore


def evaluate_20d(xx: np.ndarray) -> np.ndarray:
    """Evaluate the 20D test function from Alemazkoor & Meidani (2018).

    Parameters
    ----------
    xx : np.ndarray
        A 20-dimensional input values given by an ``(N, 20)`` array
        where ``N`` is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a one-dimensional array of length ``N``.
    """
    yy = np.sum(xx[:, :-1] * xx[:, 1:], axis=1)

    return yy


class Alemazkoor20D(UQTestFunFixDimABC):
    """An implementation of the 20D function of Alemazkoor & Meidani (2018)."""

    _tags = ["metamodeling"]
    _description = (
        "High-dimensional low-degree polynomial from Alemazkoor "
        "& Meidani (2018)"
    )
    _available_inputs = AVAILABLE_INPUTS_20D
    _available_parameters = None
    _default_input_dimension = 20

    evaluate = staticmethod(evaluate_20d)  # type: ignore



"""
Module with an implementation of the 3D test function from Moon (2010).

The Moon3D function is a three-dimensional scalar-valued function used in
[1] to illustrate the analytical derivation of Sobol' sensitivity indices.

References
----------

1. H. Moon, "Design and Analysis of Computer Experiments for Screening Input
   Variables," Ph.D. dissertation, Ohio State University, Ohio, 2010.
   URL: http://rave.ohiolink.edu/etdc/view?acc_num=osu1275422248
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Moon3D"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Moon2010": {
        "function_id": "Moon3D",
        "description": (
            "Probabilistic input model for the 3D test function "
            "from Moon (2010)"
        ),
        "marginals": [
            {
                "name": "X1",
                "distribution": "uniform",
                "parameters": [0, 1],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "uniform",
                "parameters": [0, 1],
                "description": None,
            },
            {
                "name": "X3",
                "distribution": "uniform",
                "parameters": [0, 1],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """The evaluation function for the Moon (2010) 3D function.

    Parameters
    ----------
    xx : np.ndarray
        One-dimensional input values given by an N-by-3 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = xx[:, 0] + xx[:, 1] + 3 * xx[:, 0] * xx[:, 2]

    return yy


class Moon3D(UQTestFunFixDimABC):
    """An implementation of the 3D function of Moon (2010)."""

    _tags = ["sensitivity"]
    _description = "Three-dimensional function from Moon (2010)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the damped oscillator model.

The damped oscillator model is a seven-dimensional scalar-valued function
that computes the relative displacement of a secondary spring
under a white noise base acceleration.
The model was first proposed in [1] and used in the context of reliability
analysis in, for examples, [2], [3], [4], and [5].
Note, however, that the reliability analysis variant differs
from the base model.
Used in the context of reliability analysis,
the model also includes additional parameters related to a capacity factor and
load such that the performance function can be computed.

The damped oscillator model is based on a two-degree-of-freedom
primary-secondary mechanical system characterized by two masses, two springs,
and the corresponding damping ratios.

References
----------
1. Takeru Igusa and Armen Der Kiureghian, “Dynamic characterization of
   two‐degree‐of‐freedom equipment‐structure systems,”
   Journal of Engineering Mechanics, vol. 111, no. 1, pp. 1–19, 1985.
   DOI: 10.1061/(ASCE)0733-9399(1985)111:1(1)
2. Armen Der Kiureghian and Mario De Stefano, "An efficient algorithm for
   second-order reliability analysis," Department of Civil and Environmental
   Engineering, University of California, Berkeley, UCB/SEMM-90/20, 1990.
3. Armen Der Kiureghian and Mario De Stefano, “Efficient algorithm for
   second‐order reliability analysis,” Journal of Engineering Mechanics,
   vol. 117, no. 12, pp. 2904–2923, 1991.
   DOI: 10.1061/(ASCE)0733-9399(1991)117:12(2904)
4. J.-M. Bourinet, F. Deheeger, and M. Lemaire, “Assessing small failure
   probabilities by combined subset simulation and Support Vector Machines,”
   Structural Safety, vol. 33, no. 6, pp. 343–353, 2011.
   DOI: 10.1016/j.strusafe.2011.06.001.
5. Vincent Dubourg, “Adaptive surrogate models for reliability analysis
   and reliability-based design optimization,”
   Université Blaise Pascal - Clermont II, Clermont-Ferrand, France, 2011.
   URL: https://sites.google.com/site/vincentdubourg/phd-thesis
"""

import numpy as np

from uqtestfuns.core.custom_typing import (
    ProbInputSpecs,
    FunParamSpecs,
    MarginalSpecs,
)
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC
from .utils import lognorm2norm_mean, lognorm2norm_std

__all__ = ["DampedOscillator", "DampedOscillatorReliability"]


MARGINALS_DERKIUREGHIAN1991: MarginalSpecs = [  # From [2]
    {
        "name": "Mp",
        "distribution": "lognormal",
        "parameters": [
            lognorm2norm_mean(1.5, 0.1 * 1.5),
            lognorm2norm_std(1.5, 0.1 * 1.5),
        ],
        "description": "Primary mass",
    },
    {
        "name": "Ms",
        "distribution": "lognormal",
        "parameters": [
            lognorm2norm_mean(0.01, 0.1 * 0.01),
            lognorm2norm_std(0.01, 0.1 * 0.01),
        ],
        "description": "Secondary mass",
    },
    {
        "name": "Kp",
        "distribution": "lognormal",
        "parameters": [
            lognorm2norm_mean(1.0, 0.2 * 1.0),
            lognorm2norm_std(1.0, 0.2 * 1.0),
        ],
        "description": "Primary spring stiffness",
    },
    {
        "name": "Ks",
        "distribution": "lognormal",
        "parameters": [
            lognorm2norm_mean(0.01, 0.2 * 0.01),
            lognorm2norm_std(0.01, 0.2 * 0.01),
        ],
        "description": "Secondary spring stiffness",
    },
    {
        "name": "Zeta_p",
        "distribution": "lognormal",
        "parameters": [
            lognorm2norm_mean(0.05, 0.4 * 0.05),
            lognorm2norm_std(0.05, 0.4 * 0.05),
        ],
        "description": "Primary damping ratio",
    },
    {
        "name": "Zeta_s",
        "distribution": "lognormal",
        "parameters": [
            lognorm2norm_mean(0.02, 0.5 * 0.02),
            lognorm2norm_std(0.02, 0.5 * 0.02),
        ],
        "description": "Secondary damping ratio",
    },
    {
        "name": "S0",
        "distribution": "lognormal",
        "parameters": [
            lognorm2norm_mean(100.0, 0.1 * 100.0),
            lognorm2norm_std(100.0, 0.1 * 100.0),
        ],
        "description": "White noise base acceleration",
    },
]

AVAILABLE_INPUTS_BASE: ProbInputSpecs = {
    "DerKiureghian1991": {
        "function_id": "DampedOscillator",
        "description": (
            "Probabilistic input model for the Damped Oscillator model "
            "from Der Kiureghian and De Stefano (1991)."
        ),
        "marginals": MARGINALS_DERKIUREGHIAN1991,
        "copulas": None,
    },
}


AVAILABLE_INPUTS_RELIABILITY: ProbInputSpecs = {
    "DerKiureghian1990a": {
        "function_id": "DampedOscillatorReliability",
        "description": (
            "Input model #1 for the damped oscillator reliability "
            "from Der Kiureghian and De Stefano (1990)"
        ),
        "marginals": MARGINALS_DERKIUREGHIAN1991
        + [
            {
                "name": "Fs",
                "distribution": "lognormal",
                "parameters": [
                    lognorm2norm_mean(15.0, 0.1 * 15.0),
                    lognorm2norm_std(15.0, 0.1 * 15.0),
                ],
                "description": "Force capacity of the secondary spring",
            },
        ],
        "copulas": None,
    },
    "DerKiureghian1990b": {
        "function_id": "DampedOscillatorReliability",
        "description": (
            "Input model #2 for the damped oscillator reliability "
            "from Der Kiureghian and De Stefano (1990)"
        ),
        "marginals": MARGINALS_DERKIUREGHIAN1991
        + [
            {
                "name": "Fs",
                "distribution": "lognormal",
                "parameters": [
                    lognorm2norm_mean(21.5, 0.1 * 21.5),
                    lognorm2norm_std(21.5, 0.1 * 21.5),
                ],
                "description": "Force capacity of the secondary spring",
            },
        ],
        "copulas": None,
    },
    "DerKiureghian1990c": {
        "function_id": "DampedOscillatorReliability",
        "description": (
            "Input model #3 for the damped oscillator reliability "
            "from Der Kiureghian and De Stefano (1990)"
        ),
        "marginals": MARGINALS_DERKIUREGHIAN1991
        + [
            {
                "name": "Fs",
                "distribution": "lognormal",
                "parameters": [
                    lognorm2norm_mean(27.5, 0.1 * 27.5),
                    lognorm2norm_std(27.5, 0.1 * 27.5),
                ],
                "description": "Force capacity of the secondary spring",
            },
        ],
        "copulas": None,
    },
}


AVAILABLE_PARAMETERS_RELIABILITY: FunParamSpecs = {
    "DerKiureghian1990": {
        "function_id": "DampedOscillatorReliability",
        "description": (
            "Parameter set for the damped oscillator reliability problem "
            "from Der Kiureghian and De Stefano (1990)"
        ),
        "declared_parameters": [
            {
                "keyword": "pf",
                "value": 3.0,
                "type": float,
                "description": "Peak factor [-]",
            },
        ],
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the rms displacement of the damped oscillator model.

    Parameters
    ----------
    xx : np.ndarray
        A 7-dimensional input values given by an N-by-7 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The mean-square relative displacement of the secondary spring.
    """

    # Get the parameters
    mm_p = xx[:, 0]  # Primary mass
    mm_s = xx[:, 1]  # Secondary mass
    kk_p = xx[:, 2]  # Primary spring stiffness
    kk_s = xx[:, 3]  # Secondary spring stiffness
    zt_p = xx[:, 4]  # Damping ratio of the primary damper
    zt_s = xx[:, 5]  # Damping ratio of the secondary damper
    ss_0 = xx[:, 6]  # White noise base acceleration intensity

    # Compute natural frequencies
    omega_p = np.sqrt(kk_p / mm_p)  # Primary system
    omega_s = np.sqrt(kk_s / mm_s)  # Secondary system

    # Compute additional parameters
    gamma = mm_s / mm_p  # relative mass
    omega_a = (omega_p + omega_s) / 2.0  # average natural frequency
    zt_a = (zt_p + zt_s) / 2.0  # average damping ratio
    theta = (omega_p - omega_s) / omega_a  # tuning parameter

    # Compute the mean-square relative displacement of the secondary spring
    first_term = np.pi * ss_0 / 4 / zt_s / (omega_s**3)
    second_term = (
        zt_a
        * zt_s
        / (zt_p * zt_s * (4 * zt_a**2 + theta**2) + gamma * zt_a**2)
    )
    third_term = (
        (zt_p * omega_p**3 + zt_s * omega_s**3)
        * omega_p
        / 4
        / zt_a
        / (omega_a**4)
    )

    # NOTE: This is squared displacement
    xx_s = first_term * second_term * third_term

    return np.sqrt(xx_s)


class DampedOscillator(UQTestFunFixDimABC):
    """A concrete implementation of the Damped oscillator test function."""

    _tags = ["metamodeling", "sensitivity"]
    _description = (
        "Damped oscillator model from Igusa and Der Kiureghian (1985)"
    )
    _available_inputs = AVAILABLE_INPUTS_BASE
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore


def evaluate_reliability(xx: np.ndarray, pf: float):
    """Evaluate the performance function of the system reliability.

    Parameters
    ----------
    xx : np.ndarray
        An 8-dimensional input values given by an N-by-7 array
        where N is the number of input values.
    pf : float
        The peak factor of the system.

    Returns
    -------
    np.ndarray
        The system's performance evaluation. If less than or equal to zero,
        the system is in failed state.
        The output is a 1-dimensional array of length N.
    """
    rms_disp = evaluate(xx[:, :-1])  # root-mean-square displacement
    kk_s = xx[:, 3]  # Secondary spring stiffness
    ff_s = xx[:, -1]  # Force capacity of the secondary spring

    yy = ff_s - pf * kk_s * rms_disp

    return yy


class DampedOscillatorReliability(UQTestFunFixDimABC):
    """A concrete implementation of the Damped oscillator reliability func."""

    _tags = ["reliability"]
    _description = (
        "Performance function from Der Kiureghian and De Stefano (1990)"
    )
    _available_inputs = AVAILABLE_INPUTS_RELIABILITY
    _available_parameters = AVAILABLE_PARAMETERS_RELIABILITY
    _default_input_id = "DerKiureghian1990a"
    _default_parameters_id = "DerKiureghian1990"

    evaluate = staticmethod(evaluate_reliability)  # type: ignore



"""
Module with an implementation of the Franke's test function.

The Franke's test functions are a set of six two-dimensional scalar-valued
functions. The functions were introduced in [1] in the context of
scattered data interpolation and was used in the context of metamodeling
in [2].

The six functions are:

- (1st) Franke function: Two Gaussian peaks and a Gaussian dip on a surface
  slopping down toward the upper right boundary.
- (2nd) Franke function: Two nearly flat regions joined by a sharp rise running
  diagonally.
- (3rd) Franke function: A saddle shaped surface.
- (4th) Franke function: A Gaussian hill that slopes off in a gentle fashion.
- (5th) Franke function: A steep Gaussian hill which becomes almost zero at
  the boundaries.
- (6th) Franke function: A part of a sphere.

Four of the Franke functions, namely the 2nd, 4th, 5th, and 6th are modified
from [3] (namely, S5, S3, S2 and S1, respectively).

While the Franke's original report [1] contains in total six two-dimensional
test functions, only the first Franke function is commonly known
as the "Franke function".

References
----------

1. Richard Franke, "A critical comparison of some methods for interpolation
   of scattered data," Naval Postgraduate School, Monterey, Canada,
   Technical Report No. NPS53-79-003, 1979.
   URL: https://core.ac.uk/reader/36727660
2. Ben Haaland and Peter Z. G. Qian, “Accurate emulators for large-scale
   computer experiments,” The Annals of Statistics, vol. 39, no. 6,
   pp. 2974-3002, 2011. DOI: 10.1214/11-AOS929
3. D. H. McLain, "Drawing contours from arbitrary data points," The Computer
   Journal, vol. 17, no. 4, pp. 318-324, 1974.
   DOI: 10.1093/comjnl/17.4.318
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Franke1", "Franke2", "Franke3", "Franke4", "Franke5", "Franke6"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Franke1979": {
        "function_id": "Franke",
        "description": (
            "Input specification for the test functions from Franke (1979)."
        ),
        "marginals": [
            {
                "name": "X1",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


DEFAULT_INPUT_SELECTION = "Franke1979"

COMMON_METADATA = dict(
    _tags=["metamodeling"],
    _available_inputs=AVAILABLE_INPUTS,
    _available_parameters=None,
    _description="from Franke (1979)",
)


def evaluate_franke1(xx: np.ndarray):
    """Evaluate the (1st) Franke function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the (1st) Franke function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """

    xx0 = 9 * xx[:, 0]
    xx1 = 9 * xx[:, 1]

    # Compute the (first) Franke function
    term_1 = 0.75 * np.exp(-0.25 * ((xx0 - 2) ** 2 + (xx1 - 2) ** 2))
    term_2 = 0.75 * np.exp(
        -1.00 * ((xx0 + 1) ** 2 / 49.0 + (xx1 + 1) ** 2 / 10.0)
    )
    term_3 = 0.50 * np.exp(-0.25 * ((xx0 - 7) ** 2 + (xx1 - 3) ** 2))
    term_4 = 0.20 * np.exp(-1.00 * ((xx0 - 4) ** 2 + (xx1 - 7) ** 2))

    yy = term_1 + term_2 + term_3 - term_4

    return yy


class Franke1(UQTestFunFixDimABC):
    """A concrete implementation of the (1st) Franke function.

    The function features two Gaussian peaks and a Gaussian dip.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"(1st) Franke function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_franke1)  # type: ignore


def evaluate_franke2(xx: np.ndarray):
    """Evaluate the (2nd) Franke function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the (2nd) Franke function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = (np.tanh(9 * (xx[:, 1] - xx[:, 0])) + 1) / 9.0

    return yy


class Franke2(UQTestFunFixDimABC):
    """A concrete implementation of the (2nd) Franke function.

    The function features two plateaus joined by a steep hill.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"(2nd) Franke function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_franke2)  # type: ignore


def evaluate_franke3(xx: np.ndarray):
    """Evaluate the (3rd) Franke function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the (3rd) Franke function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    term_1 = 1.25 + np.cos(5.4 * xx[:, 1])
    term_2 = 6 * (1 + (3 * xx[:, 0] - 1) ** 2)

    yy = term_1 / term_2

    return yy


class Franke3(UQTestFunFixDimABC):
    """A concrete implementation of the (3rd) Franke function.

    The function features a saddle shaped surface.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"(3rd) Franke function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_franke3)  # type: ignore


def evaluate_franke4(xx: np.ndarray):
    """Evaluate the (4th) Franke function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the (4th) Franke function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = (
        np.exp(-81.0 / 16.0 * ((xx[:, 0] - 0.5) ** 2 + (xx[:, 1] - 0.5) ** 2))
        / 3.0
    )

    return yy


class Franke4(UQTestFunFixDimABC):
    """A concrete implementation of the (4th) Franke function.

    The function features a gentle Gaussian hill.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"(4th) Franke function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_franke4)  # type: ignore


def evaluate_franke5(xx: np.ndarray):
    """Evaluate the (5th) Franke function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the (5th) Franke function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = (
        np.exp(-81.0 / 4.0 * ((xx[:, 0] - 0.5) ** 2 + (xx[:, 1] - 0.5) ** 2))
        / 3.0
    )

    return yy


class Franke5(UQTestFunFixDimABC):
    """A concrete implementation of the (5th) Franke function.

    The function features a steep Gaussian hill.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"(5th) Franke function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_franke5)  # type: ignore


def evaluate_franke6(xx: np.ndarray):
    """Evaluate the (6th) Franke function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the (6th) Franke function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = (
        np.sqrt(64 - 81 * ((xx[:, 0] - 0.5) ** 2 + (xx[:, 1] - 0.5) ** 2))
        / 9.0
        - 0.5
    )

    return yy


class Franke6(UQTestFunFixDimABC):
    """A concrete implementation of the (6th) Franke function.

    The function features a part of a sphere.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"(6th) Franke function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_franke6)  # type: ignore
